
#include <Richedit.h>

#ifndef _TEXTSERV_H
#define _TEXTSERV_H
#if _MSC_VER > 1000
#pragma once
#endif

#define S_MSG_KEY_IGNORED \
	MAKE_HRESULT(SEVERITY_SUCCESS, FACILITY_ITF, 0x201)

enum TXTBACKSTYLE {
	TXTBACK_TRANSPARENT = 0,		//@emem	background should show through
	TXTBACK_OPAQUE,					//@emem	erase background
};


enum TXTHITRESULT {
	TXTHITRESULT_NOHIT	        = 0,	//@emem	no hit
	TXTHITRESULT_TRANSPARENT	= 1,	//@emem point is within the text's rectangle, but
										//in a transparent region
	TXTHITRESULT_CLOSE	        = 2,	//@emem	point is close to the text
	TXTHITRESULT_HIT	        = 3		//@emem dead-on hit
};

enum TXTNATURALSIZE {
    TXTNS_FITTOCONTENT		= 1,		//@emem	Get a size that fits the content
    TXTNS_ROUNDTOLINE		= 2			//@emem Round to the nearest whole line.
};

enum TXTVIEW {
	TXTVIEW_ACTIVE = 0,
	TXTVIEW_INACTIVE = -1
};

enum CHANGETYPE
{
	CN_GENERIC		= 0,				//@emem Nothing special happened
	CN_TEXTCHANGED	= 1,				//@emem the text changed
	CN_NEWUNDO		= 2,				//@emem	A new undo action was added
	CN_NEWREDO		= 4					//@emem A new redo action was added
};

struct CHANGENOTIFY {
	DWORD	dwChangeType;				//@field TEXT changed, etc
	void *	pvCookieData; 				//@field cookie for the undo action
										// associated with the change.
};

#define TXTBIT_RICHTEXT			1		// rich-text control
#define TXTBIT_MULTILINE		2		// single vs multi-line control
#define TXTBIT_READONLY			4		// read only text
#define TXTBIT_SHOWACCELERATOR	8		// underline accelerator character
#define TXTBIT_USEPASSWORD		0x10	// use password char to display text
#define TXTBIT_HIDESELECTION	0x20	// show selection when inactive
#define TXTBIT_SAVESELECTION	0x40	// remember selection when inactive
#define TXTBIT_AUTOWORDSEL		0x80	// auto-word selection
#define TXTBIT_VERTICAL			0x100	// vertical
#define TXTBIT_SELBARCHANGE 	0x200	// notification that the selection bar width
										// has changed.
										// FUTURE: move this bit to the end to
										// maintain the division between
										// properties and notifications.
#define TXTBIT_WORDWRAP  		0x400	// if set, then multi-line controls
										// should wrap words to fit the available
										// display
#define	TXTBIT_ALLOWBEEP		0x800	// enable/disable beeping
#define TXTBIT_DISABLEDRAG      0x1000  // disable/enable dragging
#define TXTBIT_VIEWINSETCHANGE	0x2000	// the inset changed
#define TXTBIT_BACKSTYLECHANGE	0x4000
#define TXTBIT_MAXLENGTHCHANGE	0x8000
#define TXTBIT_SCROLLBARCHANGE	0x10000
#define TXTBIT_CHARFORMATCHANGE 0x20000
#define TXTBIT_PARAFORMATCHANGE	0x40000
#define TXTBIT_EXTENTCHANGE		0x80000
#define TXTBIT_CLIENTRECTCHANGE	0x100000	// the client rectangle changed
#define TXTBIT_USECURRENTBKG	0x200000	// tells the renderer to use the current background
											// color rather than the system default for an entire line


class ITextHost : public IUnknown
{
public:

	//@cmember Get the DC for the host
	virtual HDC 		TxGetDC() = 0;

	//@cmember Release the DC gotten from the host
	virtual INT			TxReleaseDC(HDC hdc) = 0;
	
	//@cmember Show the scroll bar
	virtual BOOL 		TxShowScrollBar(INT fnBar, BOOL fShow) = 0;

	//@cmember Enable the scroll bar
	virtual BOOL 		TxEnableScrollBar (INT fuSBFlags, INT fuArrowflags) = 0;

	//@cmember Set the scroll range
	virtual BOOL 		TxSetScrollRange(
							INT fnBar,
							LONG nMinPos,
							INT nMaxPos,
							BOOL fRedraw) = 0;

	//@cmember Set the scroll position
	virtual BOOL 		TxSetScrollPos (INT fnBar, INT nPos, BOOL fRedraw) = 0;

	//@cmember InvalidateRect
	virtual void		TxInvalidateRect(LPCRECT prc, BOOL fMode) = 0;

	//@cmember Send a WM_PAINT to the window
	virtual void 		TxViewChange(BOOL fUpdate) = 0;
	
	//@cmember Create the caret
	virtual BOOL		TxCreateCaret(HBITMAP hbmp, INT xWidth, INT yHeight) = 0;

	//@cmember Show the caret
	virtual BOOL		TxShowCaret(BOOL fShow) = 0;

	//@cmember Set the caret position
	virtual BOOL		TxSetCaretPos(INT x, INT y) = 0;

	//@cmember Create a timer with the specified timeout
	virtual BOOL 		TxSetTimer(UINT idTimer, UINT uTimeout) = 0;

	//@cmember Destroy a timer
	virtual void 		TxKillTimer(UINT idTimer) = 0;

	//@cmember Scroll the content of the specified window's client area
	virtual void		TxScrollWindowEx (
							INT dx,
							INT dy,
							LPCRECT lprcScroll,
							LPCRECT lprcClip,
							HRGN hrgnUpdate,
							LPRECT lprcUpdate,
							UINT fuScroll) = 0;
	
	//@cmember Get mouse capture
	virtual void		TxSetCapture(BOOL fCapture) = 0;

	//@cmember Set the focus to the text window
	virtual void		TxSetFocus() = 0;

	//@cmember Establish a new cursor shape
	virtual void 	TxSetCursor(HCURSOR hcur, BOOL fText) = 0;

	//@cmember Converts screen coordinates of a specified point to the client coordinates
	virtual BOOL 		TxScreenToClient (LPPOINT lppt) = 0;

	//@cmember Converts the client coordinates of a specified point to screen coordinates
	virtual BOOL		TxClientToScreen (LPPOINT lppt) = 0;

	//@cmember Request host to activate text services
	virtual HRESULT		TxActivate( LONG * plOldState ) = 0;

	//@cmember Request host to deactivate text services
   	virtual HRESULT		TxDeactivate( LONG lNewState ) = 0;

	//@cmember Retrieves the coordinates of a window's client area
	virtual HRESULT		TxGetClientRect(LPRECT prc) = 0;

	//@cmember Get the view rectangle relative to the inset
	virtual HRESULT		TxGetViewInset(LPRECT prc) = 0;

	//@cmember Get the default character format for the text
	virtual HRESULT 	TxGetCharFormat(const CHARFORMATW **ppCF ) = 0;

	//@cmember Get the default paragraph format for the text
	virtual HRESULT		TxGetParaFormat(const PARAFORMAT **ppPF) = 0;

	//@cmember Get the background color for the window
	virtual COLORREF	TxGetSysColor(int nIndex) = 0;

	//@cmember Get the background (either opaque or transparent)
	virtual HRESULT		TxGetBackStyle(TXTBACKSTYLE *pstyle) = 0;

	//@cmember Get the maximum length for the text
	virtual HRESULT		TxGetMaxLength(DWORD *plength) = 0;

	//@cmember Get the bits representing requested scroll bars for the window
	virtual HRESULT		TxGetScrollBars(DWORD *pdwScrollBar) = 0;

	//@cmember Get the character to display for password input
	virtual HRESULT		TxGetPasswordChar(TCHAR *pch) = 0;

	//@cmember Get the accelerator character
	virtual HRESULT		TxGetAcceleratorPos(LONG *pcp) = 0;

	//@cmember Get the native size
    virtual HRESULT		TxGetExtent(LPSIZEL lpExtent) = 0;

	//@cmember Notify host that default character format has changed
	virtual HRESULT 	OnTxCharFormatChange (const CHARFORMATW * pcf) = 0;

	//@cmember Notify host that default paragraph format has changed
	virtual HRESULT		OnTxParaFormatChange (const PARAFORMAT * ppf) = 0;

	//@cmember Bulk access to bit properties
	virtual HRESULT		TxGetPropertyBits(DWORD dwMask, DWORD *pdwBits) = 0;

	//@cmember Notify host of events
	virtual HRESULT		TxNotify(DWORD iNotify, void *pv) = 0;

	// Far East Methods for getting the Input Context
//#ifdef WIN95_IME
	virtual HIMC		TxImmGetContext() = 0;
	virtual void		TxImmReleaseContext( HIMC himc ) = 0;
//#endif

	//@cmember Returns HIMETRIC size of the control bar.
	virtual HRESULT		TxGetSelectionBarWidth (LONG *lSelBarWidth) = 0;

};
	
#endif // _TEXTSERV_H

GUID IID_ITextServices = { 0x8d33f740, 0xcf58, 0x11ce, {0xa8, 0x9d, 0x00, 0xaa, 0x00, 0x6c, 0xad, 0xc5}};
const IID IID_ITextHost = { 0xc5bdd8d0, 0xd26e, 0x11ce, {0xa8, 0x9e, 0x00, 0xaa, 0x00, 0x6c, 0xad, 0xc5}};

class TextHost : public ITextHost {
public:
	PARAFORMAT paraFmt_;
	CHARFORMATW charFmt_;

	TextHost() {}
	STDMETHODIMP QueryInterface( REFIID iid, void ** ppvObject ) {
		if ( iid == IID_ITextHost ) 
			*ppvObject = (void*) (ITextHost*)this;
		else if ( iid == IID_IUnknown ) 
			*ppvObject = (void*) (IUnknown*)this;
		else {
			*ppvObject = NULL;
			return E_NOINTERFACE;
		}
		return S_OK;
	}

	STDMETHODIMP_(ULONG) AddRef() {	return 0; }
	STDMETHODIMP_(ULONG) Release() { return 0; }

	HDC TxGetDC() {return NULL;}
	INT TxReleaseDC(HDC hdc) {return 0;}
	BOOL TxShowScrollBar(INT fnBar, BOOL fShow) {return FALSE;}
	BOOL TxEnableScrollBar (INT fuSBFlags, INT fuArrowflags) {return FALSE;}
	BOOL TxSetScrollRange(INT fnBar, LONG nMinPos, INT nMaxPos, BOOL fRedraw) {return FALSE;}
	BOOL TxSetScrollPos (INT fnBar, INT nPos, BOOL fRedraw) {return FALSE;}
	void TxInvalidateRect(LPCRECT prc, BOOL fMode) {}
	void TxViewChange(BOOL fUpdate) {}
	BOOL TxCreateCaret(HBITMAP hbmp, INT xWidth, INT yHeight) {return TRUE;}
	BOOL TxShowCaret(BOOL fShow) {return TRUE;}
	BOOL TxSetCaretPos(INT x, INT y) {return TRUE;}
	BOOL TxSetTimer(UINT idTimer, UINT uTimeout) {return TRUE;}
	void TxKillTimer(UINT idTimer) {}
	void TxScrollWindowEx (INT dx, INT dy, LPCRECT lprcScroll, LPCRECT lprcClip, HRGN hrgnUpdate, LPRECT lprcUpdate, UINT fuScroll) {}
	void TxSetCapture(BOOL fCapture) {}
	void TxSetFocus() {}
	void TxSetCursor(HCURSOR hcur, BOOL fText) {}
	BOOL TxScreenToClient (LPPOINT lppt) {return TRUE;}
	BOOL TxClientToScreen (LPPOINT lppt) {return TRUE;}
	HRESULT TxActivate( LONG * plOldState ) {return S_OK;}
	HRESULT TxDeactivate( LONG lNewState ){return S_OK;}
	HRESULT TxGetClientRect(LPRECT prc){return S_OK;}
	HRESULT TxGetViewInset(LPRECT prc){ prc->left = 0; prc->right = 0; prc->top = 100; prc->bottom = 55;	return S_OK; }
	COLORREF TxGetSysColor(int nIndex){return 0;}
	HRESULT TxGetScrollBars(DWORD *pdwScrollBar){return S_OK;}

	HRESULT TxGetCharFormat(const CHARFORMATW **ppCF ){
		*ppCF = &charFmt_;
		return S_OK;
	}

	HRESULT TxGetParaFormat(const PARAFORMAT **ppPF){
		*ppPF = &paraFmt_;
		return S_OK;
	}

	HRESULT TxGetBackStyle(TXTBACKSTYLE *pstyle) {
		*pstyle = TXTBACK_TRANSPARENT;
		return S_OK;
	}

	HRESULT TxGetMaxLength(DWORD *plength){
		*plength = INFINITE;
		return S_OK;
	}

	HRESULT TxGetAcceleratorPos(LONG *pcp){
		*pcp = -1;
		return S_OK;
	}

    HRESULT TxGetExtent(LPSIZEL lpExtent){return E_NOTIMPL;}
	HRESULT OnTxCharFormatChange (const CHARFORMATW * pcf){return S_OK;}
	HRESULT OnTxParaFormatChange (const PARAFORMAT * ppf){return S_OK;}
	HRESULT TxGetPasswordChar(TCHAR *pch){return S_FALSE;}

	HRESULT TxGetPropertyBits(DWORD dwMask, DWORD *pdwBits){
		*pdwBits = TXTBIT_RICHTEXT | TXTBIT_USECURRENTBKG;
		return S_OK;
	}

	HRESULT TxNotify(DWORD iNotify, void *pv){return S_OK;}
	virtual HIMC TxImmGetContext(){return NULL;}
	void TxImmReleaseContext( HIMC himc ) {}

	HRESULT TxGetSelectionBarWidth (LONG *lSelBarWidth) {
		*lSelBarWidth = 0;
		return S_OK;
	}
};












